#!/usr/bin/ruby1.8
#******************************************************************************
# Module:
#	Proxy
#
# Description:
#	Implements a network client proxy for event type messaging.
#
#	Event messages have no return type, only allow pass by value,
#	do not generate exceptions back to the client, and are
#	"asynchronous" in nature.
#
# Author:
#	John W. M. Stevens
#******************************************************************************

require 'TcpSocket'
require 'RubyKerberos'

#==============================================================================
# Class:
#	Proxy
#
# Description:
#	Convert message sends for network transport.
#==============================================================================

class Proxy < Krb5TcpClient

	#--------------------------------------------------------------------------
	# Method:
	#	initialize
	#
	# Parameters:
	#	port    - Port in interface of service to connect to.
	#	fqdn    - Interface that service listening in.
	#	service - Service name.
	#
	# Description:
	#	Initialize ActiveObject input queue.
	#
	# Returns:
	#	Initialized object.
	#--------------------------------------------------------------------------

	def initialize(
		port,
		fqdn,
		service )

		#	Initialize super class.
		super()

		#	Save address of receiver.
		@port    = port
		@fqdn    = fqdn
		@service = service

		#	Connect to service and authenticate.
		self.connect(
			@port,
			@fqdn,
			@service )
	end

	#--------------------------------------------------------------------------
	# Method:
	#	method_missing
	#
	# Parameters:
	#	message    - String representation of message.
	#	parameters - Parameters of message.
	#
	# Description:
	#	Convert message for nework transport.  This method will be
	#	invoked for every message sent to the Proxy that the Proxy
	#	object does not implement.
	#
	# Returns:
	#	Nothing.
	#--------------------------------------------------------------------------

	def method_missing(
		message,
		*parameters )

		#	Marshal message and parameters.
		payload =
			Marshal.dump(
				[ message.to_s, parameters ] )

		#	Forward message to receiver.
		self.writeMessage(
			payload )
	end
end

#==============================================================================
# Class:
#	ServiceProxy
#
# Description:
#	Convert message sends for network transport.
#==============================================================================

class ServiceProxy < Krb5TcpProxy

	#--------------------------------------------------------------------------
	# Method:
	#	initialize
	#
	# Parameters:
	#	port    - Port in interface of service to connect to.
	#	fqdn    - Interface that service listening in.
	#	service - Service name.
	#
	# Description:
	#	Initialize ActiveObject input queue.
	#
	# Returns:
	#	Initialized object.
	#--------------------------------------------------------------------------

	def initialize(
		client,
		port,
		fqdn,
		service )

		#	Initialize super class.
		super(
			client )

		#	Save address of receiver.
		@port    = port
		@fqdn    = fqdn
		@service = service

		#	Connect to service and authenticate.
		self.connect(
			@port,
			@fqdn,
			@service )
	end

	#--------------------------------------------------------------------------
	# Method:
	#	method_missing
	#
	# Parameters:
	#	message    - String representation of message.
	#	parameters - Parameters of message.
	#
	# Description:
	#	Convert message for nework transport.  This method will be
	#	invoked for every message sent to the Proxy that the Proxy
	#	object does not implement.
	#
	# Returns:
	#	Nothing.
	#--------------------------------------------------------------------------

	def method_missing(
		message,
		*parameters )

		#	Marshal message and parameters.
		payload =
			Marshal.dump(
				[ message.to_s, parameters ] )

		#	Forward message to receiver.
		self.writeMessage(
			payload )
	end
end
