#!/usr/bin/ruby1.8
#******************************************************************************
# Module:
#	NetworkInterfaces
#
# Description:
#	Query system for network interfaces.
#
# Author:
#	John W. M. Stevens
#******************************************************************************

require 'socket'

#==============================================================================
# Class:
#	IPv4Address
#
# Description:
#	Query DNS system to forward and reverse resolve address
#	and interface name.
#==============================================================================

class IPv4Address

	#--------------------------------------------------------------------------
	# Method:
	#	initialize
	#
	# Parameters:
	#	label       - Interface label (lo, eth0, etc).
	#	ipv4Address - IPv4 Address.
	#	broadcast   - IPv4 Broadcast Address.
	#	netmask     - IPv4 Network mask.
	#
	# Description:
	#	Initialize an IPv4 address object.
	#
	# Returns:
	#	Nothing.
	#--------------------------------------------------------------------------

	def initialize(
		label,
		ipv4Address,
		broadcast,
		netmask )

		#	Save label
		@label = label

		#	Resolve address.
		addr =
			Socket.gethostbyname(
				ipv4Address )
		@ipv4Address = addr[ 0 ]

		#	Reverse resolve address to DNS.
		addr =
			Socket.gethostbyaddr(
				addr[3], addr[2] )
		@interfaceName = addr[ 0 ]

		#	Save broadcast and netmask.
		@broadcast = broadcast
		@netmask   = netmask
	end

	#--------------------------------------------------------------------------
	# Method:
	#	socketAddress
	#
	# Parameters:
	#	port - The port number of the socket address.
	#
	# Description:
	#	Pack the interface and port number information into an internet
	#	protocol version 4 socket address structure and return it.
	#
	# Returns:
	#	The socket address structure packed into a string.
	#--------------------------------------------------------------------------

	def socketAddress(
		port )

		return
			Socket.pack_sockaddr_in(
				port,
				@interfaceName )
	end

	#--------------------------------------------------------------------------
	# Method:
	#	to_s
	#
	# Parameters:
	#	None
	#
	# Description:
	#	Return a string representation of this address.
	#
	# Returns:
	#	Return a string representation of this address.
	#--------------------------------------------------------------------------

	def to_s
		return "#{@label} - #{@interfaceName}: " +
			"#{@ipv4Address} / #{@broadcast} / #{@netmask}"
	end
end

#==============================================================================
# Class:
#	Loopback
#
# Description:
#	Query DNS system to forward and reverse resolve address
#	and interface name.
#==============================================================================

class Loopback < IPv4Address

	#--------------------------------------------------------------------------
	# Method:
	#	initialize
	#
	# Parameters:
	#	label       - Interface label (lo, eth0, etc).
	#	ipv4Address - IPv4 Address.
	#	broadcast   - IPv4 Broadcast Address.
	#	netmask     - IPv4 Network mask.
	#
	# Description:
	#	Initialize an IPv4 address object.
	#
	# Returns:
	#	Nothing.
	#--------------------------------------------------------------------------

	def initialize(
		label,
		ipv4Address,
		netmask )

		#	Resolve address.
		super(
			label,
			ipv4Address,
			nil,
			netmask )

		#	The ipv4Address must be 127.0.0.1 and the interface name
		#	must be localhost.
		if @ipv4Address != '127.0.0.1' or @interfaceName != 'localhost'
		then
			raise Exception, "Local interface not properly configured!"
		end
	end
end

#==============================================================================
# Class:
#	NetworkInterfaces
#
# Description:
#	Query for all network interfaces in computer.
#==============================================================================

class NetworkInterfaces < Hash

	#--------------------------------------------------------------------------
	# Method:
	#	getAllInterfaces
	#
	# Parameters:
	#	None.
	#
	# Description:
	#	Parse the output of ifconfig.
	#
	# Returns:
	#	None
	#--------------------------------------------------------------------------

	def getAllInterfaces(
		p )

		#	For all interfaces, do:
		while not p.eof?
			#	Load list with interfaces that have
			#	internet addresses configured on them.
			while not p.eof?
				#	Get next line.
				line = p.readline

				#	Get lines, discard until an interface
				#	label is seen.
				break if line =~ /^[^\s]/
			end

			#	End of file already?
			if p.eof? or not line
				$stderr.print "ifconfig: unexpected end of output.\n"
				break
			end

			#	Get label.
			list  = line.split
			label = list[ 0 ]

			#	Split on white space.
			if list[ 1 ] != 'Link'
				$stderr.print "ifconfig: output parse error.\n"
				$stderr.print "   expected 'Link' in line.\n"
				$stderr.print "   #{line}\n"
				break
			end

			#	Get link type.
			words    = list[ 2 ].split( ':' )
			linkType = words[ 1 ]
			if words[ 0 ] != 'encap'       ||
			   linkType   != 'Ethernet' &&
			   linkType   != 'Local'
			then
				$stderr.print "ifconfig unknown link type.\n"
				$stderr.print "   expected 'encap' in line.\n"
				$stderr.print "   Link Type: (#{linkType}).\n"
				$stderr.print "   #{line}\n"
				break
			end

			#	Get next line.
			if not p.eof?
				line = p.readline
			else
				$stderr.print "ifconfig: unexpected end of output.\n"
				break
			end

			#	Is this an IP interface?
			if line !~ /inet addr:/
				#	Loop until end of file, or blank line.
				while not p.eof?
					line = p.readline
					break if line =~ /^\s*$/
				end

				#	Next loop iteration.
				next
			end

			#	Collect information.
			case linkType
			when 'Ethernet' then
				#	Split line into words.
				list = line.split

				#	Get IPv4 address.
				words = list[ 1 ].split( ':' )
				if words[ 0 ] == 'addr'
					ipv4Address = words[ 1 ]
				else
					$stderr.print "ifconfig: no address.\n"
					$stderr.print "   expected 'addr' in field 1.\n"
					$stderr.print "   #{line}\n"
					ipv4Address = nil
				end

				#	Get broadcast address.
				words = list[ 2 ].split( ':' )
				if words[ 0 ] == 'Bcast'
					broadcast = words[ 1 ]
				else
					$stderr.print "ifconfig: no broadcast.\n"
					$stderr.print "   expected 'Bcast' in field 2.\n"
					$stderr.print "   #{line}\n"
					broadcast = nil
				end

				#	Get netmask.
				words = list[ 3 ].split( ':' )
				if words[ 0 ] == 'Mask'
					netmask = words[ 1 ]
				else
					$stderr.print "ifconfig: no netmask.\n"
					$stderr.print "   expected 'Netmask' in field 3.\n"
					$stderr.print "   #{line}\n"
					netmask = nil
				end

				#	Create new address object.
				begin
					interface =
						IPv4Address.new(
							label,
							ipv4Address,
							broadcast,
							netmask )
				rescue
					$stderr.print "DNS problems!\n"
					$stderr.print "   IPv4Address - #{$!}\n"
					exit( 1 )
				end
			when 'Local' then
				#	Split line into words.
				list = line.split

				#	Get IPv4 address.
				words = list[ 1 ].split( ':' )
				if words[ 0 ] == 'addr'
					ipv4Address = words[ 1 ]
				else
					$stderr.print "ifconfig: no address.\n"
					$stderr.print "   expected 'addr' in field 1.\n"
					$stderr.print "   #{line}\n"
					ipv4Address = nil
				end

				#	Get netmask.
				words = list[ 2 ].split( ':' )
				if words[ 0 ] == 'Mask'
					netmask = words[ 1 ]
				else
					$stderr.print "ifconfig: no netmask.\n"
					$stderr.print "   expected 'Netmask' in field 2.\n"
					$stderr.print "   #{line}\n"
					netmask = nil
				end

				#	Create new loopback address object.
				begin
					interface =
						Loopback.new(
							label,
							ipv4Address,
							netmask )
				rescue
					$stderr.print "DNS problems!\n"
					$stderr.print "   Loopback - #{$!}\n"
					exit( 1 )
				end
			else
				$stderr.print "ifconfig: unknown link type.\n"
			end

			#	Add to interfaces.
			self[ label ] = interface

			#	Loop until end of file, or blank line.
			while not p.eof?
				line = p.readline
				break if line =~ /^\s*$/
			end
		end
	end

	#--------------------------------------------------------------------------
	# Method:
	#	initialize
	#
	# Parameters:
	#	None.
	#
	# Description:
	#	Query for all network interfaces, map from label to
	#	network information information.
	#
	# Returns:
	#	Nothing.
	#--------------------------------------------------------------------------

	def initialize
		#	Execute query to get list of active ("up") interfaces.
		IO.popen( "/sbin/ifconfig", "r" ) { |p|
			getAllInterfaces(
				p )
		}
	end
end
