#!/usr/bin/ruby1.8
#******************************************************************************
# Module:
#	Receptionist
#
# Description:
#	Manage socket connections.
#
# Author:
#	John W. M. Stevens
#******************************************************************************

require	'TcpSocket'
require	'RubyKerberos'

class Receptionist < Krb5TcpServer

	#--------------------------------------------------------------------------
	# Method:
	#	initialize
	#
	# Parameters:
	#	port              - Port to listen for connections on.
	#	interface         - Interface containing port to listen on.
	#	activeObjectKlass - Class of active object to proxy for.
	#
	# Description:
	#	Initialize for connection listening.
	#
	# Returns:
	#	Initialized object.
	#--------------------------------------------------------------------------

	def initialize(
		port,
		interface,
		activeObjectKlass )

		#	Create message queue.
		@goNow       = true
		@connections = Array.new

		#	Build pipe, create active object.
		readChannel, writeChannel = IO.pipe
		activeObject =
			activeObjectKlass.new(
				readChannel )

		#	Start active object.
		@childPid = fork do
			#	Close write channel.
			writeChannel.close

			#	Execute active object.
			activeObject.processMessagesUntilStopped

			#	Exit.
			exit 0
		end

		#	Close read channel of pipe, save write channel.
		readChannel.close
		@channel = writeChannel

		#	Save data attributes.
		@port        = port
		@interface   = interface
		@serviceName = activeObject.class.name

		#	Initialize the Kerberized TCP Server, passing it the class
		#	to use to create new objects on accept.
		super(
			Krb5TcpConnection )

		#	Listen on port on given local interface.
		listenOn(
			@port,
			@interface,
			@serviceName )

		#	Save the connections.
		@connections << self
	end

	#--------------------------------------------------------------------------
	# Method:
	#	stopMeBeforeIGoAroundAgain
	#
	# Parameters:
	#	None.
	#
	# Description:
	#	Set stop processing flag.
	#
	# Returns:
	#	None.
	#--------------------------------------------------------------------------

	def stopMeBeforeIGoAroundAgain
		@goNow = false
	end

	#--------------------------------------------------------------------------
	# Method:
	#	processMessagesUntilStopped
	#
	# Parameters:
	#	None.
	#
	# Description:
	#	Execute messages in queue in first in, first out order.  All
	#	messages must be event type messages.  To enforce this, all
	#	return values and side effects will be disposed of.
	#
	#	If the message queue is empty, wait for a message to be
	#	queued.
	#
	# Returns:
	#	None.
	#--------------------------------------------------------------------------

	def processMessagesUntilStopped
		#	Accept connections.
		while @goNow
			#	Get read sets.
			readSet =
				TcpSocket.select(
					@connections,
					nil,
					nil )[ 0 ]

			#	Process read events.
			readSet.each { |proxy|
				#	Is this self?
				if proxy == self
					#	Yes.  Acquire, and authenticate.
					client = proxy.accept
					client.authenticate
					@connections << client
				else
					begin
						#	No.  Read message, write it to
						#	the pipe.
						message = proxy.readMessage

						#	Pipe to object.
						@channel.write(
							[ message.length ].pack( 'l' ) )
						@channel.write(
							message )
					rescue
						#	Remove client socket.
						@connections.delete(
							proxy )
					end
				end
			}
		end

		#	Wait for active object to quit.
		Process.waitpid( @childPid )
	end
end
