/******************************************************************************
* Module:
*	RubyKerberos
*
* Description:
*	Kerberized TCP Socket classes.
*
* Author:
*	John W. M. Stevens
******************************************************************************/

#include	<stdio.h>
#include	<stdlib.h>
#include	<errno.h>
#include	<limits.h>
#include	<string.h>
#include	<stdint.h>
#include	<netdb.h>
#include	<unistd.h>

#include	"ruby.h"
#include	"intern.h"

#include	"RubyKerberos.h"

/*	TcpSocket base class reference.	*/
extern	VALUE	cTcpSocket;

/*	Declare Kerberized class.	*/
		VALUE	cKrb5TcpSocket;
		VALUE	cKrb5TcpServer;
		VALUE	cKrb5TcpClient;
		VALUE	cKrb5TcpConnection;
		VALUE	cKrb5TcpProxy;
static	VALUE	eKrb5TcpSocket;
static	VALUE	eKrb5TcpServer;
static	VALUE	eKrb5TcpClient;
static	VALUE	eKrb5TcpConnection;
static	VALUE	eKrb5TcpProxy;

static	char	*proxyCacheName   = 0;
static	char	*keyTabsDirectory = "KRB5_SERVER_KEYTABS";
static	char	serverKeyTabs[ PATH_MAX + 1 ];

#define	SAFE_FREE( a )	\
	( (a) = (void *) (((a) != 0) ? free( (a) ), 0 : 0 ))

/*-----------------------------------------------------------------------------
| Method:
|	Krb5_Error
|
| Parameters:
|	error  - Kerberos 5 error code.
|	format - Error message format string.
|	...	   - Values to print.
|
| Description:
|	Print Kerberos 5 error message and associated message data.
|
| Returns:
|	None.
|
| Raises:
|	None.
-----------------------------------------------------------------------------*/

static
void
Krb5_Error(
	krb5_error_code	error,
	char			*format,
	... )
{
	auto	va_list		args;

	/*	Start variable arguments.	*/
    va_start(
		args,
		format );
    vfprintf(
		stderr,
		format,
		args );
    va_end(
		args );

	/*	Print message.	*/
    fprintf(
		stderr,
		"%s\n",
		error_message(
			error ) );
}

/*-----------------------------------------------------------------------------
| Method:
|	Krb5_FreeServerContext
|
| Parameters:
|	serverContext - Server context structure whose data should be freed.
|
| Description:
|	Discard any data set in the Kerberos 5 server context structure.
|
| Returns:
|	None.
|
| Raises:
|	None.
-----------------------------------------------------------------------------*/

static
void
Krb5_FreeServerContext(
	Krb5ServerContext_t	*serverContext )
{
}

/*-----------------------------------------------------------------------------
| Method:
|	SetLocalAddress
|
| Parameters:
|	sock				 - Socket file number.
|	authorizationContext - Authorization context to set remote addresses
|						   into for use by krb5_rd_priv and krb5_mk_priv.
|	serverContext		 - Server context structure.
|	serviceName			 - Name of service for error messages.
|	exception            - Exception to raise.
|
| Description:
|	Get local address and set it into the authorization context for
|	use by encryption routines.
|
| Returns:
|	None.
|
| Raises:
|	None.
-----------------------------------------------------------------------------*/

static
void
SetLocalAddress(
	int					sock,
	krb5_auth_context	authorizationContext,
	Krb5ServerContext_t	*serverContext,
	char				*serviceName,
	VALUE				exception )
{
	auto	int					sizeOf;
	auto	struct sockaddr_in	localAddress;
	auto	krb5_address		address;
	auto	krb5_error_code		code;

	/*	Can we get the remote address?	*/
	sizeOf = sizeof( localAddress );
	memset(
		&localAddress,
		0,
		sizeOf );
	if (
		getsockname(
			sock,
			(struct sockaddr *) &localAddress,
			&sizeOf ) < 0 )
	{
		/*	No.  Raise exception.	*/
		rb_raise(
			exception,
			"SetLocalAddress (%s:%d) - Can't get local address,\n"
			"   %s\n",
			__FILE__,
			__LINE__,
			strerror(
				errno ));
	}

	/*	Set remote address into authorization context.	*/
	address.addrtype = ADDRTYPE_INET;
	address.length   = sizeof( localAddress.sin_addr );
	address.contents = (krb5_octet *) &localAddress.sin_addr;
	if ( (code =
		krb5_auth_con_setaddrs(
			serverContext->context,
			authorizationContext,
			&address,
			0 )) != 0)
	{
		/*	Communicate error.	*/
		com_err(
			serviceName,
			code,
			"\n   Return from krb5_auth_con_setaddrs().");

		/*	Raise exception.	*/
		rb_raise(
			exception,
			"SetLocalAddress (%s:%d) - Failed to set local\n"
			"   address into authorization context.\n",
			__FILE__,
			__LINE__ );
	}

	/*	Set remote port into authorization context.	*/
	address.addrtype = ADDRTYPE_IPPORT;
	address.length   = sizeof( localAddress.sin_port );
	address.contents = (krb5_octet *) &localAddress.sin_port;
	if ( (code =
		krb5_auth_con_setports(
			serverContext->context,
			authorizationContext,
			&address,
			0 )) != 0)
	{
		/*	Communicate error.	*/
		com_err(
			serviceName,
			code,
			"\n   Return from krb5_auth_con_setports().");

		/*	Raise exception.	*/
		rb_raise(
			exception,
			"SetLocalAddress (%s:%d) - Failed to set local\n"
			"   port into authorization context.\n",
			__FILE__,
			__LINE__ );
	}
}

/*-----------------------------------------------------------------------------
| Method:
|	SetRemoteAddress
|
| Parameters:
|	sock				 - Socket file number.
|	authorizationContext - Authorization context to set remote addresses
|						   into for use by krb5_rd_priv and krb5_mk_priv.
|	serverContext		 - Server context structure.
|	serviceName			 - Name of service for error messages.
|	exception            - Exception to raise.
|
| Description:
|	Get remote address and set it into the authorization context for
|	use by encryption routines.
|
| Returns:
|	None.
|
| Raises:
|	None.
-----------------------------------------------------------------------------*/

static
void
SetRemoteAddress(
	int					sock,
	krb5_auth_context	authorizationContext,
	Krb5ServerContext_t	*serverContext,
	char				*serviceName,
	VALUE				exception )
{
	auto	int					sizeOf;
	auto	struct sockaddr_in	remoteAddress;
	auto	krb5_address		address;
	auto	krb5_error_code		code;

	/*	Can we get the remote address?	*/
	sizeOf = sizeof( remoteAddress );
	memset(
		&remoteAddress,
		0,
		sizeOf );
	if (
		getpeername(
			sock,
			(struct sockaddr *) &remoteAddress,
			&sizeOf ) < 0 )
	{
		/*	No.  Raise exception.	*/
		rb_raise(
			exception,
			"SetRemoteAddress (%s:%d) - Can't get remote address,\n"
			"   %s\n",
			__FILE__,
			__LINE__,
			strerror(
				errno ));
	}

	/*	Set remote address into authorization context.	*/
	address.addrtype = ADDRTYPE_INET;
	address.length   = sizeof( remoteAddress.sin_addr );
	address.contents = (krb5_octet *) &remoteAddress.sin_addr;
	if ( (code =
		krb5_auth_con_setaddrs(
			serverContext->context,
			authorizationContext,
			0,
			&address )) != 0)
	{
		/*	Communicate error.	*/
		com_err(
			serviceName,
			code,
			"\n   Return from krb5_auth_con_setaddrs().");

		/*	Raise exception.	*/
		rb_raise(
			exception,
			"SetRemoteAddress (%s:%d) - Failed to set remote\n"
			"   address into authorization context.\n",
			__FILE__,
			__LINE__ );
	}

	/*	Set remote port into authorization context.	*/
	address.addrtype = ADDRTYPE_IPPORT;
	address.length   = sizeof( remoteAddress.sin_port );
	address.contents = (krb5_octet *) &remoteAddress.sin_port;
	if ( (code =
		krb5_auth_con_setports(
			serverContext->context,
			authorizationContext,
			0,
			&address )) != 0)
	{
		/*	Communicate error.	*/
		com_err(
			serviceName,
			code,
			"\n   Return from krb5_auth_con_setports().");

		/*	Raise exception.	*/
		rb_raise(
			exception,
			"SetRemoteAddress (%s:%d) - Failed to set remote\n"
			"   port into authorization context.\n",
			__FILE__,
			__LINE__ );
	}
}

/*-----------------------------------------------------------------------------
| Method:
|	CreateReplayCache
|
| Parameters:
|	sock				 - Socket file number.
|	authorizationContext - Authorization context to set remote addresses
|						   into for use by krb5_rd_priv and krb5_mk_priv.
|	serverContext		 - Server context structure.
|	exception            - Exception to raise.
|
| Description:
|	Create and set a replay cache into the authorization context.
|
| Returns:
|	None.
|
| Raises:
|	None.
-----------------------------------------------------------------------------*/

static
void
CreateReplayCache(
	int					sock,
	krb5_auth_context	authorizationContext,
	Krb5ServerContext_t	*serverContext,
	VALUE				exception )
{
	auto	int					sizeOf;
	auto	struct sockaddr_in	localAddress;
	auto	krb5_address		address;
	auto	krb5_rcache			replayCache;
	auto	krb5_data			replayCacheName;
	auto	char				*string;
	auto	krb5_error_code		code;

	/*	Can we get the local address?	*/
	sizeOf = sizeof( localAddress );
	memset(
		&localAddress,
		0,
		sizeOf );
	if (
		getsockname(
			sock,
			(struct sockaddr *) &localAddress,
			&sizeOf ) < 0 )
	{
		/*	No.  Raise exception.	*/
		rb_raise(
			exception,
			"CreateReplayCache (%s:%d) - Can't get local address,\n"
			"   %s\n",
			__FILE__,
			__LINE__,
			strerror(
				errno ));
	}

	/*	Generate replay cache name.	*/
	address.addrtype = ADDRTYPE_IPPORT;
	address.length   = sizeof( localAddress.sin_port );
	address.contents = (krb5_octet *) &localAddress.sin_port;
	if ( (code =
		krb5_gen_replay_name(
			serverContext->context,
			&address,
			"_lcl_cch",
			&string )) != 0 )
	{
		/*	No.  Raise exception.	*/
		rb_raise(
			exception,
			"CreateReplayCache (%s:%d) - Can't generate replay"
			"cache name.\n",
			__FILE__,
			__LINE__ );
	}

	/*	Set replay cache name string.	*/
	replayCacheName.data   = string;
	replayCacheName.length =
		strlen(
			string );

	/*	Get replay cache.	*/
	if ( (code =
		krb5_get_server_rcache(
			serverContext->context,
			&replayCacheName,
			&replayCache )) != 0 )
	{
		/*	No.  Raise exception.	*/
		rb_raise(
			exception,
			"CreateReplayCache (%s:%d) - Can't get replay cache.\n",
			__FILE__,
			__LINE__ );
	}

	/*	Set replay cache into authorization context.	*/
	krb5_auth_con_setrcache(
		serverContext->context,
		authorizationContext,
		replayCache );
}

/*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@ Class:
@	Krb5TcpSocket
@
@ Description:
@	Kerberos 5 TCP Connection.
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@*/

/*-----------------------------------------------------------------------------
| Method:
|	Free
|
| Parameters:
|	dataAttributes - The data attributes to free.
|
| Description:
|	Free the Kerberos 5 data associated with the object.
|
| Returns:
|	None.
|
| Raises:
|	None.
-----------------------------------------------------------------------------*/

static
void
Krb5TcpSocket_Free(
	void	*dataAttributes )
{
	auto	Krb5TcpSocketDAs_t	*das;
	auto	Krb5ServerContext_t		*sc;

	/*	Get Kerberos 5 data structure pointer.	*/
	das = (Krb5TcpSocketDAs_t *) dataAttributes;
	sc  = &das->serverContext;

	/*	Discard Kerberos5 stuff.	*/
	if ( das->authorizationContext )
		krb5_auth_con_free(
			sc->context,
			das->authorizationContext );
	if ( das->clientPrincipal )
		krb5_free_principal(
			sc->context,
			das->clientPrincipal );

	/*	Discard strings.	*/
	if ( das->clientInterface )
		free(
			das->clientInterface );

	/*	If the socket is still open, close it.	*/
	if (das->sock >= 0)
	{
		/*	Closet the socket.	*/
		close(
			das->sock );
	}

	/*	Zero structure fields.	*/
	memset(
		das,
		0,
		sizeof( Krb5TcpSocketDAs_t ));
}

/*-----------------------------------------------------------------------------
| Method:
|	Allocate
|
| Parameters:
|	klass - The pointer to this Ruby object's class.
|
| Description:
|	Allocate a new Krb5TcpSocket object.
|
| Returns:
|	Returns newly allocated object.
|
| Raises:
|	Nothing.
-----------------------------------------------------------------------------*/

static
VALUE
Krb5TcpSocket_Allocate(
	VALUE	klass )
{
	auto	VALUE				object;
	auto	Krb5TcpSocketDAs_t	*das;

	/*	Allocate kerberos 5 data structure.	*/
	object =
		Data_Make_Struct(
			klass,
			Krb5TcpSocketDAs_t,
			0,
			Krb5TcpSocket_Free,
			das );
	das->sock = -1;

	/*	Return new object.	*/
	return object;
}

/*-----------------------------------------------------------------------------
| Method:
|	Authenticate
|
| Parameters:
|	self - The pointer to the this Ruby object.
|
| Description:
|	Authenticate client that connected to this server.
|
| Returns:
|	Returns self.
|
| Raises:
|	ExceptionKrb5TcpSocket on authentication failure.
-----------------------------------------------------------------------------*/

static
VALUE
Krb5TcpSocket_Authenticate(
	VALUE	self)
{
	auto	char					buffer[ 10 ];
	auto	krb5_error_code			authError;
	auto	Krb5TcpSocketDAs_t		*das;
	auto	Krb5ServerContext_t		*sc;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		Krb5TcpSocketDAs_t,
		das );
	sc = &das->serverContext;

	/*	Receive authorization.	*/
	if ( (authError =
		krb5_recvauth(
			sc->context,
			&das->authorizationContext,
			(krb5_pointer) &das->sock,
			"1",
			sc->serverPrincipal,
			0,
			sc->keyTable,
			NULL )) != 0)
	{
		/*	Set the authorization flag.	*/
		das->krb5AuthFlag = 0;

		/*	Raise an exception.	*/
		rb_raise(
			eKrb5TcpSocket,
			"Krb5TcpSocket_Authenticate (%s:%d) - Kerberos 5\n"
			"   authentication failed",
			__FILE__,
			__LINE__ );
	}

	/*	Set the authorization flag.	*/
	das->krb5AuthFlag = 1;

	/*	Return self.	*/
	return self;
}

/*-----------------------------------------------------------------------------
| Method:
|	WriteMessage
|
| Parameters:
|	self - The pointer to the this Ruby object.
|
| Description:
|	Close the listening socket.  Allow the free routine to clean up the
|	Kerberos 5 data.
|
| Returns:
|	Returns nil.
|
| Raises:
|	Nothing.
-----------------------------------------------------------------------------*/

static
VALUE
Krb5TcpSocket_WriteMessage(
	VALUE	self,
	VALUE	message )
{
	auto	Krb5TcpSocketDAs_t	*das;
	auto	Krb5ServerContext_t	*sc;
	auto	long				length;
	auto	krb5_error_code		code;
	auto	krb5_data			data;
	auto	krb5_data			packet;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		Krb5TcpSocketDAs_t,
		das );
	sc = &das->serverContext;

	/*	Get buffer to write.	*/
	data.data =
		rb_str2cstr(
			message,
			&length );
	data.length = length;

	/*	Can we encrypt the message?	*/
	if ( (code =
		krb5_mk_priv(
			das->serverContext.context,
			das->authorizationContext,
			&data,
			&packet,
			0 )) != 0 )
	{
		/*	No.  Raise exception to indicate encryption error.	*/
		rb_raise(
			eKrb5TcpSocket,
			"Krb5TcpSocket_WriteMessage (%s:%d) - Encryption error,\n"
			"   %s\n",
			__FILE__,
			__LINE__,
			error_message(
				code ) );
	}

	/*	Can we write message length to socket?	*/
	length =
		htonl(
			(long) packet.length );
	if (
		write(
			das->sock,
			&length,
			sizeof( length )) != sizeof( length ) )
	{
		/*	No.  Raise exception to indicate write error.	*/
		free(
			packet.data );
		rb_raise(
			eKrb5TcpSocket,
			"Krb5TcpSocket_WriteMessage (%s:%d) - Socket write error,\n"
			"   %s",
			__FILE__,
			__LINE__,
			strerror(
				errno ) );
	}

	/*	Can we write encrypted data to socket?	*/
	if (
		write(
			das->sock,
			packet.data,
			packet.length ) != packet.length )
	{
		/*	No.  Raise exception to indicate write error.	*/
		free(
			packet.data );
		rb_raise(
			eKrb5TcpSocket,
			"Krb5TcpSocket_WriteMessage (%s:%d) - Socket write error,\n"
			"   [%d] %s",
			__FILE__,
			__LINE__,
			errno,
			strerror(
				errno ) );
	}

	/*	Free encrypted buffer.	*/
	length = packet.length;
	free(
		packet.data );

	/*	Return number of bytes written to socket.	*/
	return
		INT2NUM(
			length );
}

/*-----------------------------------------------------------------------------
| Method:
|	ReadMessage
|
| Parameters:
|	self - The pointer to the this Ruby object.
|
| Description:
|	Close the listening socket.  Allow the free routine to clean up the
|	Kerberos 5 data.
|
| Returns:
|	Returns nil.
|
| Raises:
|	Nothing.
-----------------------------------------------------------------------------*/

static
VALUE
Krb5TcpSocket_ReadMessage(
	VALUE	self )
{
	auto	Krb5TcpSocketDAs_t	*das;
	auto	Krb5ServerContext_t	*sc;
	auto	krb5_error_code		code;
	auto	krb5_data			data;
	auto	krb5_data			packet;
	auto	long				netLength;
	auto	long				length;
	auto	int					size;
	auto	VALUE				message;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		Krb5TcpSocketDAs_t,
		das );
	sc = &das->serverContext;

	/*	Can we read the message length?	*/
	if ( (size =
		read(
			das->sock,
			&netLength,
			sizeof( netLength ) )) != sizeof( netLength ))
	{
		/*	No.  Raise exception.	*/
		rb_raise(
			eKrb5TcpSocket,
			"Krb5TcpSocket_ReadMessage (%s:%d) - Socket read error,\n"
			"   %s",
			__FILE__,
			__LINE__,
			strerror(
				errno ) );
	}
	else if (size == 0)
	{
		/*	End of file read error.	*/
		rb_raise(
			eKrb5TcpSocket,
			"Krb5TcpSocket_ReadMessage (%s:%d) - No message to read,\n"
			"   %s",
			__FILE__,
			__LINE__,
			strerror(
				errno ) );
	}

	/*	Convert from network order to host order.	*/
	length =
		ntohl(
			netLength );

	/*	Can we allocate a buffer?	*/
	packet.length = length;
	if ( (packet.data =
		calloc(
			1,
			length )) == 0)
	{
		/*	No.  Raise  memory allocation error.	*/
		rb_raise(
			eKrb5TcpSocket,
			"Krb5TcpSocket_ReadMessage (%s:%d) - Encryption error,\n"
			"   can't allocate packet data buffer.\n",
			__FILE__,
			__LINE__ );
	}

	/*	Can we read the message?	*/
	if ( (size =
		read(
			das->sock,
			packet.data,
			length )) < 0)
	{
		/*	No.  Free buffer and raise read error.	*/
		free(
			packet.data );
		rb_raise(
			eKrb5TcpSocket,
			"Krb5TcpSocket_ReadMessage (%s:%d) - Socket read error,\n"
			"   %s",
			__FILE__,
			__LINE__,
			strerror(
				errno ) );
	}

	/*	Did we read as many bytes as we needed to?	*/
	if (size != packet.length)
	{
		/*	No.  Free buffer and raise read error.	*/
		free(
			packet.data );
		rb_raise(
			eKrb5TcpSocket,
			"Krb5TcpSocket_ReadMessage (%s:%d) - Short read error,\n"
			"   %s",
			__FILE__,
			__LINE__,
			strerror(
				errno ) );
	}

	/*	Can we decrypt the message?	*/
	data.data   = 0;
	data.length = 0;
	if (
		krb5_rd_priv(
			sc->context,
			das->authorizationContext,
			&packet,
			&data,
			0 ) != 0 )
	{
		/*	No.  Raise exception to indicate encryption error.	*/
		free(
			packet.data );
		rb_raise(
			eKrb5TcpSocket,
			"Krb5TcpSocket_ReadMessage (%s:%d) - Encryption error,\n"
			"   %s\n",
			__FILE__,
			__LINE__,
			error_message(
				code ) );
	}

	/*	Return message.	*/
	free(
		packet.data );
	message =
		rb_str_new(
			data.data,
			data.length );
	free(
		data.data );
	return message;
}

/*-----------------------------------------------------------------------------
| Method:
|	Close
|
| Parameters:
|	self - The pointer to the this Ruby object.
|
| Description:
|	Close the listening socket.  Allow the free routine to clean up the
|	Kerberos 5 data.
|
| Returns:
|	Returns nil.
|
| Raises:
|	Accept errors.
-----------------------------------------------------------------------------*/

static
VALUE
Krb5TcpSocket_Close(
	VALUE	self)
{
	auto	Krb5TcpSocketDAs_t	*das;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		Krb5TcpSocketDAs_t,
		das );

	/*	Close socket.	*/
	if (das->sock >= 0)
	{
		/*	Close the socket.	*/
		close(
			das->sock );
		das->sock = -1;
	}

	/*	Return nil.	*/
	return Qnil;
}

static
void
Initialize_Krb5TcpSocket(
	void )
{
	/*	Create Krb5TcpSocket class.	*/
	cKrb5TcpSocket =
		rb_define_class(
			"Krb5TcpSocket",
			cTcpSocket );

	/*	Add allocation method.	*/
	rb_define_alloc_func(
		cKrb5TcpSocket,
		Krb5TcpSocket_Allocate );

	/*	Define methods in class.	*/
	rb_define_method(
		cKrb5TcpSocket,
		"authenticate",
		Krb5TcpSocket_Authenticate,
		0 );
	rb_define_method(
		cKrb5TcpSocket,
		"close",
		Krb5TcpSocket_Close,
		0 );
	rb_define_method(
		cKrb5TcpSocket,
		"readMessage",
		Krb5TcpSocket_ReadMessage,
		0 );
	rb_define_method(
		cKrb5TcpSocket,
		"writeMessage",
		Krb5TcpSocket_WriteMessage,
		1 );

	/*	Create exception class.	*/
	eKrb5TcpSocket =
		rb_define_class(
			"ExceptionKrb5TcpSocket",
			rb_eStandardError );
}

/*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@ Class:
@	Krb5TcpServer
@
@ Description:
@	Kerberos 5 TCP Server.
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@*/

/*-----------------------------------------------------------------------------
| Method:
|	Free
|
| Parameters:
|	dataAttributes - The data attributes to free.
|
| Description:
|	Free the Kerberos 5 data associated with the object.
|
| Returns:
|	None.
|
| Raises:
|	None.
-----------------------------------------------------------------------------*/

static
void
Krb5TcpServer_Free(
	void	*dataAttributes )
{
	auto	Krb5TcpServerDAs_t		*das;
	auto	Krb5ServerContext_t		*sc;

	/*	Get Kerberos 5 data structure pointer.	*/
	das = (Krb5TcpServerDAs_t *) dataAttributes;
	sc  = &das->serverContext;

	/*	Discard service strings.	*/
	if ( das->serviceInterface )
		free(
			das->serviceInterface );
	if ( das->serviceName )
		free(
			das->serviceName );

	/*	Discard Kerberos5 stuff.	*/
	if ( sc->serverPrincipal )
		krb5_free_principal(
			sc->context,
			sc->serverPrincipal );
	if ( sc->defaultCache )
		krb5_cc_close(
			sc->context,
			sc->defaultCache );
	if ( sc->keyTable )
		krb5_kt_close(
			sc->context,
			sc->keyTable );
	if ( sc->context )
		krb5_free_context(
			sc->context );

	/*	Zero data attribute fields.	*/
	das->authorizationContext = 0;
	sc->keyTable              = 0;
	sc->defaultCache          = 0;
	sc->serverPrincipal       = 0;
	sc->context               = 0;

	/*	If the socket is still open, close it.	*/
	if (das->sock >= 0)
	{
		/*	Closet the socket.	*/
		close(
			das->sock );
	}

	/*	Zero structure fields.	*/
	memset(
		das,
		0,
		sizeof( Krb5TcpServerDAs_t ));
}

/*-----------------------------------------------------------------------------
| Method:
|	Allocate
|
| Parameters:
|	klass - The pointer to this Ruby object's class.
|
| Description:
|	Allocate the Krb5TcpServer object.
|
| Returns:
|	Returns newly allocated object.
|
| Raises:
|	Accept errors.
-----------------------------------------------------------------------------*/

static
VALUE
Krb5TcpServer_Allocate(
	VALUE	klass )
{
	auto	VALUE				object;
	auto	Krb5TcpServerDAs_t	*das;

	/*	Allocate kerberos 5 data structure.	*/
	object =
		Data_Make_Struct(
			klass,
			Krb5TcpServerDAs_t,
			0,
			Krb5TcpServer_Free,
			das );
	das->sock = -1;

	/*	Return new object.	*/
	return object;
}

/*-----------------------------------------------------------------------------
| Method:
|	Accept
|
| Parameters:
|	argc - The number of method parameters.
|	argv - The method parameter list.
|	self - The pointer to the this Ruby object.
|
| Description:
|	Accept connections, create new objects that implement an initializer
|	that takes a socket descriptor number and an internet address
|	structure packed into a string.
|
| Returns:
|	Returns the new object created to wrap the accept results.
|
| Raises:
|	Accept errors.
-----------------------------------------------------------------------------*/

static
VALUE
Krb5TcpServer_Accept(
	int		argc,
	VALUE	*argv,
	VALUE	self )
{
	auto	VALUE				rNewSocket;
	auto	Krb5TcpServerDAs_t	*das;
	auto	Krb5ServerContext_t	*sc;
	auto	VALUE				values[ 2 ];

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		Krb5TcpServerDAs_t,
		das );
	sc = &das->serverContext;

	/*	Pass instance of server context to super
	*	class for use in initialization.
	*/
	values[ 0 ] =
		rb_str_new(
			(void *) sc,
			sizeof( das->serverContext ));
	rNewSocket  =
		rb_call_super(
			1,
			values );

	/*	Return new socket object.	*/
	return rNewSocket;
}

/*-----------------------------------------------------------------------------
| Method:
|	ListenOn
|
| Parameters:
|	self         - The pointer to the this Ruby object.
|	rPort        - Port number to listen on.
|	rInterface   - If present, specifies the local interface to bind to.
|                  if of zero length, specifies any interface
|	rServiceName - Name of service.
|
| Description:
|	Bind to a given port and address and start listening for connections.
|
|	By convention, the service name should be the class name of the
|	ActiveObject sub-class that this is a listener for.
|
|	The service interface string should be the Kerberos 5 service name
|	string as registered in the Kerberos system.
|
| Returns:
|	Returns the new TcpSocket like object.
|
| Raises:
|	ExceptionKrb5TcpServer on context initialization, key table and
|	principal resolution errors.
-----------------------------------------------------------------------------*/

static
VALUE
Krb5TcpServer_ListenOn(
	VALUE	self,
	VALUE	rPort,
	VALUE	rInterface,
	VALUE	rServiceName )
{
	auto	VALUE				rNewSocket;
	auto	Krb5TcpServerDAs_t	*das;
	auto	Krb5ServerContext_t	*sc;
	auto	long				length;
	auto	char				keyTableFileName[ PATH_MAX ];
	auto	VALUE				values[ 3 ];

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		Krb5TcpServerDAs_t,
		das );

	/*	Extract service port.	*/
	sc = &das->serverContext;
	das->servicePort =
		NUM2INT(
			rPort );

	/*	Get local interface DNS name.	*/
	if ( (das->serviceInterface =
		strdup(
			rb_str2cstr(
				rInterface,
				&length )) ) == 0 )
	{
		rb_raise(
			eKrb5TcpServer,
			"Krb5TcpServer_ListenOn (%s:%d) - Out of memory!",
			__FILE__,
			__LINE__ );
	}

	/*	Get service name.	*/
	if ( (das->serviceName =
		strdup(
			rb_str2cstr(
				rServiceName,
				&length )) ) == 0)
	{
		/*	Raise exception.	*/
		rb_raise(
			eKrb5TcpServer,
			"Krb5TcpServer_ListenOn (%s:%d) - Out of memory!",
			__FILE__,
			__LINE__ );
	}

	/*	Get the kerberos 5 context.	*/
	if (
		krb5_init_context(
			&sc->context ) )
	{
		/*	Raise exception.	*/
		rb_raise(
			eKrb5TcpServer,
			"Krb5TcpServer_ListenOn (%s:%d) - Init Kerberos 5 context.",
			__FILE__,
			__LINE__ );
	}

	/*	Resolve key table.	*/
	snprintf(
		keyTableFileName,
		PATH_MAX,
		"%s/%s.keytab",
		serverKeyTabs,
		das->serviceName );
	if (
		krb5_kt_resolve(
			sc->context,
			keyTableFileName,
			&sc->keyTable) )
	{
		/*	Raise exception.	*/
		rb_raise(
			eKrb5TcpServer,
			"Krb5TcpServer_ListenOn (%s:%d) - Bad key table file name:\n"
			"   (%s)",
			__FILE__,
			__LINE__,
			keyTableFileName );
	}

	/*	Resolve service name.	*/
	if (
		krb5_sname_to_principal(
			sc->context,
			das->serviceInterface,
			das->serviceName,
			KRB5_NT_UNKNOWN,
			&sc->serverPrincipal ) ) 
	{
		/*	Raise exception.	*/
		rb_raise(
			eKrb5TcpServer,
			"Krb5TcpServer_ListenOn (%s:%d) - Bad service principal:\n"
			"   (%s/%s)",
			__FILE__,
			__LINE__,
			das->serviceName,
			das->serviceInterface );
	}

	/*	Initialize super class.	*/
	values[ 0 ] = rPort;
	values[ 1 ] = rInterface;
	values[ 2 ] = rServiceName;
	return
		rb_call_super(
			3,
			values );
}

static
void
Initialize_Krb5TcpServer(
	void )
{
	auto	char	*directory;

	/*	Create TcpSocket class.	*/
	cKrb5TcpServer =
		rb_define_class(
			"Krb5TcpServer",
			cTcpSocket );

	/*	Add allocation method.	*/
	rb_define_alloc_func(
		cKrb5TcpServer,
		Krb5TcpServer_Allocate );

	/*	Define methods in class.	*/
	rb_define_method(
		cKrb5TcpServer,
		"accept",
		Krb5TcpServer_Accept,
		-1 );
	rb_define_method(
		cKrb5TcpServer,
		"listenOn",
		Krb5TcpServer_ListenOn,
		3 );

	/*	Create exception class.	*/
	eKrb5TcpServer =
		rb_define_class(
			"ExceptionKrb5TcpServer",
			rb_eStandardError );

	/*	Get server keytab directory from environment.	*/
	if ( (directory =
		getenv(
			keyTabsDirectory )) == 0 )
	{
		/*	Raise exception.	*/
		rb_raise(
			eKrb5TcpServer,
			"Initialize_Krb5TcpServer (%s:%d) - Missing "
			"KRB5_SERVER_KEYTABS\n   environmental variable",
			__FILE__,
			__LINE__ );
	}

	/*	Store server keytab directory path.	*/
	if (
		realpath(
			directory,
			serverKeyTabs ) == 0 )
	{
		/*	Raise exception.	*/
		rb_raise(
			eKrb5TcpServer,
			"Initialize_Krb5TcpServer (%s:%d) - Cannot canonicalize "
			"directory in\n   KRB5_SERVER_KEYTABS environmental variable.",
			__FILE__,
			__LINE__ );
	}

	/*	Check that the directory exists.	*/
	if (
		access(
			serverKeyTabs,
			R_OK | X_OK ) == -1)
	{
		/*	Raise exception.	*/
		rb_raise(
			eKrb5TcpServer,
			"Initialize_Krb5TcpServer (%s:%d) - Cannot access directory\n"
			"   (%s) in:\n"
			"   KRB5_SERVER_KEYTABS environmental variable.",
			__FILE__,
			__LINE__ );
	}
}

/*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@ Class:
@	Krb5TcpClient
@
@ Description:
@	Kerberos 5 TCP Client.  Use this in clients to connect to Kerberized
@	servers.
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@*/

/*-----------------------------------------------------------------------------
| Method:
|	Free
|
| Parameters:
|	dataAttributes - The data attributes to free.
|
| Description:
|	Free the Kerberos 5 data associated with the object.
|
| Returns:
|	None.
|
| Raises:
|	None.
-----------------------------------------------------------------------------*/

static
void
Krb5TcpClient_Free(
	void	*dataAttributes )
{
	auto	Krb5TcpClientDAs_t		*das;
	auto	Krb5ServerContext_t		*sc;

	/*	Get Kerberos 5 data structure pointer.	*/
	das = (Krb5TcpClientDAs_t *) dataAttributes;
	sc  = &das->serverContext;

	/*	Discard client strings.	*/
	if ( das->clientInterface )
		free(
			das->clientInterface );

	/*	Discard Kerberos5 stuff.	*/
	if ( das->clientPrincipal )
		krb5_free_principal(
			sc->context,
			das->clientPrincipal );
	if ( das->authorizationContext )
		krb5_auth_con_free(
			sc->context,
			das->authorizationContext );

	/*	Discard service strings.	*/
	if ( das->serviceInterface )
		free(
			das->serviceInterface );
	if ( das->serviceName )
		free(
			das->serviceName );

	/*	Discard Kerberos5 stuff.	*/
	if ( sc->serverPrincipal )
		krb5_free_principal(
			sc->context,
			sc->serverPrincipal );
	if ( sc->keyTable )
		krb5_kt_close(
			sc->context,
			sc->keyTable );
	if ( sc->defaultCache )
		krb5_cc_close(
			sc->context,
			sc->defaultCache );
	if ( sc->context )
		krb5_free_context(
			sc->context );

	/*	If the socket is still open, close it.	*/
	if (das->sock >= 0)
	{
		/*	Closet the socket.	*/
		close(
			das->sock );
	}

	/*	Zero structure fields.	*/
	memset(
		das,
		0,
		sizeof( Krb5TcpClientDAs_t ));
}

/*-----------------------------------------------------------------------------
| Method:
|	Allocate
|
| Parameters:
|	klass - The pointer to this Ruby object's class.
|
| Description:
|	Allocate a Krb5TcpClient object.
|
| Returns:
|	Returns newly allocated object.
|
| Raises:
|	Nothing.
-----------------------------------------------------------------------------*/

static
VALUE
Krb5TcpClient_Allocate(
	VALUE	klass )
{
	auto	VALUE				object;
	auto	Krb5TcpClientDAs_t	*das;

	/*	Allocate kerberos 5 data structure.	*/
	object =
		Data_Make_Struct(
			klass,
			Krb5TcpClientDAs_t,
			0,
			Krb5TcpClient_Free,
			das );
	das->sock = -1;

	/*	Return new object.	*/
	return object;
}

/*-----------------------------------------------------------------------------
| Method:
|	Initialize
|
| Parameters:
|	self - The pointer to the this Ruby object.
|
| Description:
|	Initialize the Krb5TcpClient object.
|
|	By convention, the service name should be the class name of the
|	ActiveObject sub-class that this is a listener for.
|
|	The service interface string should be the Kerberos service name
|	(principal name).
|
| Returns:
|	Returns super result.
|
| Raises:
|	ExceptionKrb5TcpClient on context initialization, cache and
|	principal resolution errors.
-----------------------------------------------------------------------------*/

static
VALUE
Krb5TcpClient_Initialize(
	VALUE	self )
{
	auto	Krb5TcpClientDAs_t		*das;
	auto	Krb5ServerContext_t		*sc;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		Krb5TcpClientDAs_t,
		das );
	sc = &das->serverContext;

	/*	Get the kerberos 5 context.	*/
	if (
		krb5_init_context(
			&sc->context ) )
	{
		/*	Raise exception.	*/
		sc->context = 0;
		rb_raise(
			eKrb5TcpClient,
			"Krb5TcpClient_Initialize (%s:%d) - Cannot initialize\n"
			"   Kerberos 5 Context",
			__FILE__,
			__LINE__ );
	}

	/*	Get default cache.	*/
	if (
		krb5_cc_default(
			sc->context,
			&sc->defaultCache ) )
	{
		/*	Free service principal and context.	*/
		krb5_free_context(
			sc->context );

		/*	Zero pointers.	*/
		sc->context         = 0;
		sc->defaultCache    = 0;

		/*	Raise exception.	*/
		rb_raise(
			eKrb5TcpClient,
			"Krb5TcpClient_Initialize (%s:%d) - Bad default\n"
			"   Kerberos 5 credential cache",
			__FILE__,
			__LINE__ );
	}

	/*	Get client principal.	*/
	if (
		krb5_cc_get_principal(
			sc->context,
			sc->defaultCache,
			&das->clientPrincipal) )
	{
		/*	Free Kerberos 5 stuff.	*/
		krb5_cc_close(
			sc->context,
			sc->defaultCache );
		krb5_free_context(
			sc->context );

		/*	Zero Kerberos 5 stuff.	*/
		sc->context          = 0;
		sc->defaultCache     = 0;
		das->clientPrincipal = 0;

		/*	Raise exception.	*/
		rb_raise(
			eKrb5TcpClient,
			"Krb5TcpClient_Initialize (%s:%d) - Bad Kerberos 5\n"
			"   client principal",
			__FILE__,
			__LINE__ );
	}

	/*	Invoke super class initialization.	*/
	return
		rb_call_super(
			0,
			0 );
}

/*-----------------------------------------------------------------------------
| Method:
|	Connect
|
| Parameters:
|	self         - The pointer to the this Ruby object.
|	rPort        - Port to connect to.
|	rInterface   - DNS name of interface of port to connect to.
|	rServiceName - Service name string.
|
| Description:
|	Connect this Krb5TcpClient socket to a service.
|
| Returns:
|	Returns zero.
|
| Raises:
|	ExceptionKrb5TcpClient on service name resolution and
|	authentication errors.
-----------------------------------------------------------------------------*/

static
VALUE
Krb5TcpClient_Connect(
	VALUE	self,
	VALUE	rPort,
	VALUE	rInterface,
	VALUE	rServiceName )
{
	auto	Krb5TcpClientDAs_t	*das;
	auto	Krb5ServerContext_t	*sc;
	auto	long				length;
	auto	int					error;
	auto	krb5_error			*errorReturn;
	auto	krb5_error_code		code;
	auto	VALUE				object;
	auto	krb5_data			data;
	auto	VALUE				values[ 3 ];

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		Krb5TcpClientDAs_t,
		das );
	sc = &das->serverContext;

	/*	Initialize the service name and interface
	*	strings, and save the service port number.
	*/
	das->servicePort      =
		NUM2INT(
			rPort );
	das->serviceInterface =
		strdup(
			rb_str2cstr(
				rInterface,
				&length ));
	das->serviceName      =
		strdup(
			rb_str2cstr(
				rServiceName,
				&length ));

	/*	Resolve service name.	*/
	if (
		krb5_sname_to_principal(
			sc->context,
			das->serviceInterface,
			das->serviceName,
			KRB5_NT_SRV_HST,
			&sc->serverPrincipal ) ) 
	{
		/*	Raise exception.	*/
		rb_raise(
			eKrb5TcpClient,
			"Krb5TcpClient_Connect (%s:%d) - Bad Kerberos 5\n"
			"   service principal (%s/%s)",
			__FILE__,
			__LINE__,
			das->serviceName,
			das->serviceInterface );
	}

	/*	Forward to super class.	*/
	values[ 0 ] = rPort;
	values[ 1 ] = rInterface;
	values[ 2 ] = rServiceName;
	object =
		rb_call_super(
			3,
			values );

	/*	Send some data?	*/
    data.data   = das->serviceName;
    data.length = strlen( das->serviceName );

	/*	Send authentication.	*/
	if ( (code =
		krb5_sendauth(
			sc->context,
			&das->authorizationContext,
			(krb5_pointer) &das->sock,
			"1",
			das->clientPrincipal,
			sc->serverPrincipal,
			AP_OPTS_MUTUAL_REQUIRED,
			&data,
			NULL,
			sc->defaultCache,
			&errorReturn,
			NULL,
			NULL )) != 0)
	{
		/*	Communicate error.	*/
		com_err(
			das->serviceName,
			code,
			"\n   Return from krb5_sendauth().");

		/*	Discard Kerberos 5 stuff.	*/
		if ( errorReturn )
			krb5_free_error(
				sc->context,
				errorReturn );
		if ( sc->serverPrincipal )
			krb5_free_principal(
				sc->context,
				sc->serverPrincipal );

		/*	Zero all fields.	*/
		sc->serverPrincipal = 0;
		das->krb5AuthFlag   = 0;

		/*	Raise authentication exception.	*/
		rb_raise(
			eKrb5TcpClient,
			"Krb5TcpClient_Connect (%s:%d) - Failed Kerberos 5\n"
			"   authentication to service",
			__FILE__,
			__LINE__,
			das->serviceName,
			das->serviceInterface );
	}

	/*	Set the authorization flag.	*/
	das->krb5AuthFlag = 1;

	/*	Set remote addresses into authorization context.	*/
	SetLocalAddress(
		das->sock,
		das->authorizationContext,
		sc,
		das->serviceName,
		eKrb5TcpClient );

	/*	Set remote addresses into authorization context.	*/
	CreateReplayCache(
		das->sock,
		das->authorizationContext,
		sc,
		eKrb5TcpClient );

	/*	Return zero.	*/
	return
		INT2NUM(
			0 );
}

/*-----------------------------------------------------------------------------
| Method:
|	Close
|
| Parameters:
|	self - The pointer to the this Ruby object.
|
| Description:
|	Close the listening socket.  Allow the free routine to clean up the
|	Kerberos 5 data.
|
| Returns:
|	Returns nil.
|
| Raises:
|	Nothing.
-----------------------------------------------------------------------------*/

static
VALUE
Krb5TcpClient_Close(
	VALUE	self)
{
	auto	Krb5TcpClientDAs_t	*das;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		Krb5TcpClientDAs_t,
		das );

	/*	Close socket.	*/
	if (das->sock >= 0)
	{
		/*	Close the socket.	*/
		close(
			das->sock );
		das->sock = -1;
	}

	/*	Return nil.	*/
	return Qnil;
}

static
void
Initialize_Krb5TcpClient(
	void )
{
	/*	Create Krb5TcpClient class.	*/
	cKrb5TcpClient =
		rb_define_class(
			"Krb5TcpClient",
			cKrb5TcpSocket );

	/*	Add allocation method.	*/
	rb_define_alloc_func(
		cKrb5TcpClient,
		Krb5TcpClient_Allocate );

	/*	Define methods in class.	*/
	rb_define_method(
		cKrb5TcpClient,
		"initialize",
		Krb5TcpClient_Initialize,
		0 );
	rb_define_method(
		cKrb5TcpClient,
		"connect",
		Krb5TcpClient_Connect,
		3 );
	rb_define_method(
		cKrb5TcpClient,
		"close",
		Krb5TcpClient_Close,
		0 );

	/*	Create exception class.	*/
	eKrb5TcpClient =
		rb_define_class(
			"ExceptionKrb5TcpClient",
			rb_eStandardError );
}

/*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@ Class:
@	Krb5TcpConnection
@
@ Description:
@	Kerberos 5 TCP Connection.
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@*/

/*-----------------------------------------------------------------------------
| Method:
|	Free
|
| Parameters:
|	dataAttributes - The data attributes to free.
|
| Description:
|	Free the Kerberos 5 data associated with the object.
|
| Returns:
|	None.
|
| Raises:
|	None.
-----------------------------------------------------------------------------*/

static
void
Krb5TcpConnection_Free(
	void	*dataAttributes )
{
	auto	Krb5TcpConnectionDAs_t	*das;
	auto	Krb5ServerContext_t		*sc;

	/*	Get Kerberos 5 data structure pointer.	*/
	das = (Krb5TcpConnectionDAs_t *) dataAttributes;
	sc  = &das->serverContext;

	/*	Discard Kerberos5 stuff.	*/
	if ( das->authorizationContext )
		krb5_auth_con_free(
			sc->context,
			das->authorizationContext );
	if ( das->clientPrincipal )
		krb5_free_principal(
			sc->context,
			das->clientPrincipal );

	/*	Discard strings.	*/
	if ( das->clientInterface )
		free(
			das->clientInterface );

	/*	If the socket is still open, close it.	*/
	if (das->sock >= 0)
	{
		/*	Closet the socket.	*/
		close(
			das->sock );
	}

	/*	Zero structure fields.	*/
	memset(
		das,
		0,
		sizeof( Krb5TcpConnectionDAs_t ));
}

/*-----------------------------------------------------------------------------
| Method:
|	Allocate
|
| Parameters:
|	klass - The pointer to this Ruby object's class.
|
| Description:
|	Allocate a new Krb5TcpConnection object.
|
| Returns:
|	Returns newly allocated object.
|
| Raises:
|	Nothing.
-----------------------------------------------------------------------------*/

static
VALUE
Krb5TcpConnection_Allocate(
	VALUE	klass )
{
	auto	VALUE				object;
	auto	Krb5TcpConnectionDAs_t	*das;

	/*	Allocate kerberos 5 data structure.	*/
	object =
		Data_Make_Struct(
			klass,
			Krb5TcpConnectionDAs_t,
			0,
			Krb5TcpConnection_Free,
			das );
	das->sock = -1;

	/*	Return new object.	*/
	return object;
}

/*-----------------------------------------------------------------------------
| Method:
|	Initialize
|
| Parameters:
|	argc - The number of method parameters.
|	argv - The method parameter list.
|	self - The pointer to the this Ruby object.
|
| Description:
|	Initialize the Krb5TcpConnection object.
|
|	By convention, the service name should be the class name of the
|	ActiveObject sub-class that this is a listener for.
|
|	The service string should be the Kerberos 5 principal.
|
| Returns:
|	Returns super result.
|
| Raises:
|	Accept errors.
-----------------------------------------------------------------------------*/

static
VALUE
Krb5TcpConnection_Initialize(
	int		argc,
	VALUE	*argv,
	VALUE	self )
{
	auto	Krb5TcpConnectionDAs_t	*das;
	auto	Krb5ServerContext_t		*sc;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		Krb5TcpConnectionDAs_t,
		das );
	sc = &das->serverContext;

	/*	Initialize the socket file descriptor
	*	data attribute, and the local address.
	*/
	das->sock =
		NUM2INT(
			argv[ 0 ] );

	/*	Is the address of the right length?	*/
	if (RSTRING( argv[ 1 ] )->len == sizeof( das->serviceAddress ))
		/*	Yes.  Save the internet address structure.	*/
		memcpy(
			&das->serviceAddress,
			RSTRING( argv[ 1 ] )->ptr,
			sizeof( das->serviceAddress ));
	else
		rb_raise(
			eKrb5TcpConnection,
			"Krb5TcpConnection_Initialize (%s:%d) - Address\n"
			"   wrong length (%d)",
			__FILE__,
			__LINE__,
			RSTRING( argv[ 1 ] )->len );

	/*	Is the server context of the right length?	*/
	if (RSTRING( argv[ 2 ] )->len == sizeof( das->serverContext ))
		/*	Yes.  Save the server context.	*/
		memcpy(
			sc,
			RSTRING( argv[ 2 ] )->ptr,
			sizeof( das->serverContext ));
	else
		rb_raise(
			eKrb5TcpConnection,
			"Krb5TcpConnection_Initialize (%s:%d) - Server context\n"
			"   wrong length (%d)",
			__FILE__,
			__LINE__,
			RSTRING( argv[ 2 ] )->len );

	/*	Send super.	*/
	return
		rb_call_super(
			2,
			argv );
}

/*-----------------------------------------------------------------------------
| Method:
|	Authenticate
|
| Parameters:
|	self - The pointer to the this Ruby object.
|
| Description:
|	Authenticate client that connected to this server.
|
| Returns:
|	Returns self.
|
| Raises:
|	ExceptionKrb5TcpConnection on authentication failure.
-----------------------------------------------------------------------------*/

static
VALUE
Krb5TcpConnection_Authenticate(
	VALUE	self)
{
	auto	char					buffer[ 10 ];
	auto	krb5_error_code			authError;
	auto	Krb5TcpConnectionDAs_t	*das;
	auto	Krb5ServerContext_t		*sc;
	auto	krb5_error_code			code;
	auto	krb5_address			address;
	auto	krb5_rcache				replayCache;
	auto	krb5_data				replayCacheName;
	auto	int						sizeOf;
	auto	struct sockaddr_in		localAddress;
	auto	struct sockaddr_in		remoteAddress;
	auto	char					*string;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		Krb5TcpConnectionDAs_t,
		das );
	sc = &das->serverContext;

	/*	Receive authorization.	*/
	if ( (authError =
		krb5_recvauth(
			sc->context,
			&das->authorizationContext,
			(krb5_pointer) &das->sock,
			"1",
			sc->serverPrincipal,
			0,
			sc->keyTable,
			NULL )) != 0)
	{
		/*	Set the authorization flag.	*/
		das->krb5AuthFlag = 0;

		/*	Communicate error.	*/
		com_err(
			das->serviceName,
			authError,
			"\n   Return from krb5_recvauth().");

		/*	Raise an exception.	*/
		rb_raise(
			eKrb5TcpConnection,
			"Krb5TcpConnection_Authenticate (%s:%d) - Kerberos 5\n"
			"   authentication failed\n",
			__FILE__,
			__LINE__ );
	}

	/*	Set the authorization flag.	*/
	das->krb5AuthFlag = 1;

	/*	Set remote addresses into authorization context.	*/
	SetRemoteAddress(
		das->sock,
		das->authorizationContext,
		sc,
		das->serviceName,
		eKrb5TcpConnection );

	/*	Create replay cache.	*/
	CreateReplayCache(
		das->sock,
		das->authorizationContext,
		sc,
		eKrb5TcpConnection );

	/*	Return self.	*/
	return self;
}

/*-----------------------------------------------------------------------------
| Method:
|	Close
|
| Parameters:
|	self - The pointer to the this Ruby object.
|
| Description:
|	Close the listening socket.  Allow the free routine to clean up the
|	Kerberos 5 data.
|
| Returns:
|	Returns nil.
|
| Raises:
|	Accept errors.
-----------------------------------------------------------------------------*/

static
VALUE
Krb5TcpConnection_Close(
	VALUE	self)
{
	auto	Krb5TcpConnectionDAs_t	*das;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		Krb5TcpConnectionDAs_t,
		das );

	/*	Close socket.	*/
	if (das->sock >= 0)
	{
		/*	Close the socket.	*/
		close(
			das->sock );
		das->sock = -1;
	}

	/*	Return nil.	*/
	return Qnil;
}

static
void
Initialize_Krb5TcpConnection(
	void )
{
	/*	Create Krb5TcpConnection class.	*/
	cKrb5TcpConnection =
		rb_define_class(
			"Krb5TcpConnection",
			cKrb5TcpSocket );

	/*	Add allocation method.	*/
	rb_define_alloc_func(
		cKrb5TcpConnection,
		Krb5TcpConnection_Allocate );

	/*	Define methods in class.	*/
	rb_define_method(
		cKrb5TcpConnection,
		"initialize",
		Krb5TcpConnection_Initialize,
		-1 );
	rb_define_method(
		cKrb5TcpConnection,
		"authenticate",
		Krb5TcpConnection_Authenticate,
		0 );
	rb_define_method(
		cKrb5TcpConnection,
		"close",
		Krb5TcpConnection_Close,
		0 );

	/*	Create exception class.	*/
	eKrb5TcpConnection =
		rb_define_class(
			"ExceptionKrb5TcpConnection",
			rb_eStandardError );
}

/*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@ Class:
@	Krb5TcpProxy
@
@ Description:
@	Kerberos 5 TCP Server Proxy.  Use this to connect to
@	Kerberized services.
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@*/

/*-----------------------------------------------------------------------------
| Method:
|	GetFirstPrincipalFromKeyTable
|
| Parameters:
|	das           - The data attributes for the Kerberized proxy.
|	serverContext - Server context.
|	serviceName   - The base service name.
|
| Description:
|	Get first principal from the key table.
|
| Returns:
|	Returns first service principal in key table.
|
| Raises:
|	ExceptionKrb5TcpProxy on server keytab resolution, keytab traversal
|	errors or principal parse errors.
-----------------------------------------------------------------------------*/

static
char	*
GetFirstPrincipalFromKeyTable(
	Krb5TcpProxyDAs_t	*das,
	Krb5ServerContext_t	*serverContext,
	char				*serviceName )
{
    auto	krb5_kt_cursor		cursor;
    auto	krb5_keytab_entry	entry;
	auto	char				keyTableFileName[ PATH_MAX + 1 ];
	auto	int					error;
	auto	char				*clientName;
	auto	char				*end;

	/*	Resolve key table.	*/
	snprintf(
		keyTableFileName,
		PATH_MAX,
		"%s/%s.keytab",
		serverKeyTabs,
		serviceName );
	if (
		krb5_kt_resolve(
			serverContext->context,
			keyTableFileName,
			&serverContext->keyTable) )
	{
		/*	Raise exception.	*/
		rb_raise(
			eKrb5TcpProxy,
			"GetFirstPrincipalFromKeyTable (%s:%d) - Bad key table\n"
			"   file name (%s)",
			__FILE__,
			__LINE__,
			keyTableFileName );
	}

	/*	Start sequential traversal of key tab.	*/
	if ( (error =
		krb5_kt_start_seq_get(
			serverContext->context,
			serverContext->keyTable,
			&cursor )) == 0)
	{
		/*	Get first enty in key tab.	*/
		if ( (error =
			krb5_kt_next_entry(
				serverContext->context,
				serverContext->keyTable,
				&entry,
				&cursor )) == 0)
		{
			/*	Unparse first name found.	*/
			if (
				krb5_unparse_name(
					serverContext->context,
					entry.principal,
					&clientName ) != 0)
			{
				/*	End sequential traversal of keytab.	*/
				krb5_kt_end_seq_get(
					serverContext->context,
					serverContext->keyTable,
					&cursor );

				/*	Raise exception.	*/
				rb_raise(
					eKrb5TcpProxy,
					"GetFirstPrincipalFromKeyTable (%s:%d) - Could not\n"
					"   unparse principal (%s)",
					__FILE__,
					__LINE__,
					entry.principal );
			}

			/*	Save principal.	*/
			krb5_copy_principal(
				serverContext->context,
				entry.principal,
				&das->clientPrincipal );

			/*	Discard key tab entry.	*/
			krb5_kt_free_entry(
				serverContext->context,
				&entry );
		}
		else
		{
			/*	End sequential traversal of keytab.	*/
			krb5_kt_end_seq_get(
				serverContext->context,
				serverContext->keyTable,
				&cursor );

			/*	Raise exception.	*/
			rb_raise(
				eKrb5TcpProxy,
				"GetFirstPrincipalFromKeyTable (%s:%d) - Could not\n"
				"   traverse keytab (%s)",
				__FILE__,
				__LINE__,
				keyTableFileName );
		}

		/*	End sequential traversal of keytab.	*/
		krb5_kt_end_seq_get(
			serverContext->context,
			serverContext->keyTable,
			&cursor );
    }
	else
	{
		/*	Raise exception.	*/
		rb_raise(
			eKrb5TcpProxy,
			"GetFirstPrincipalFromKeyTable (%s:%d) - Could not\n"
			"   iterate over keytab (%s)",
			__FILE__,
			__LINE__,
			keyTableFileName );
	}

	/*	Return service name.	*/
	return clientName;
}

/*-----------------------------------------------------------------------------
| Method:
|	CreateTicketCache
|
| Parameters:
|	das - Service data attribute structure pointer.
|
| Description:
|	Create ticket cache in memory.
|
| Returns:
|	Create an in-memory ticket cache.
|
| Raises:
|	ExceptionKrb5TcpProxy on out of memory errors, and credential cache
|	resolution errors.
-----------------------------------------------------------------------------*/

static
void
CreateTicketCache(
	Krb5TcpProxyDAs_t	*das )
{
	auto	int						cacheFile;
	auto	Krb5ServerContext_t		*sc;
	auto	krb5_error_code			error;

	/*	Get server context.	*/
	sc = &das->serverContext;

	/*	Has the proxy cache been created yet?	*/
	if (proxyCacheName == 0)
	{
		/*	No.  Can we allocate space for it's name?	*/
		if ( (proxyCacheName =
			calloc(
				1,
				128 )) == 0)
		{
			/*	Raise ticket cache resolution error.	*/
			rb_raise(
				eKrb5TcpProxy,
				"CreateTicketCache (%s:%d) - Out of memory!",
				__FILE__,
				__LINE__ );
		}

		/*	Create name.	*/
		sprintf(
			proxyCacheName,
			"MEMORY:%s_%d_%d",
			das->clientName,
			(int) getuid(),
			(int) getpid() );

		/*	Resolve ticket cache.	*/
		if ( (error =
			krb5_cc_resolve(
				sc->context,
				proxyCacheName,
				&sc->defaultCache )) != 0)
		{
			/*	Raise ticket cache resolution error.	*/
			rb_raise(
				eKrb5TcpProxy,
				"CreateTicketCache (%s:%d) - %s!",
				__FILE__,
				__LINE__,
				strerror(
					errno ) );
		}

		/*	Initialize ticket cache.	*/
		if ( (error =
			krb5_cc_initialize(
				sc->context,
				sc->defaultCache,
				das->clientPrincipal )) != 0)
		{
			/*	Raise ticket cache initialization error.	*/
			Krb5_Error(
				error,
				"krb5_cc_initialize" );
			rb_raise(
				eKrb5TcpProxy,
				"CreateTicketCache (%s %d) - Failed ticket\n"
				"   cache initialization!",
				__FILE__,
				__LINE__ );
		}
	}
	/*	Resolve ticket cache.	*/
	else if ( (error =
		krb5_cc_resolve(
			sc->context,
			proxyCacheName,
			&sc->defaultCache )) != 0)
	{
		/*	Raise ticket cache resolution error.	*/
		rb_raise(
			eKrb5TcpProxy,
			"CreateTicketCache (%s %d) - %s!",
			__FILE__,
			__LINE__,
			strerror(
				errno ) );
	}
}

/*-----------------------------------------------------------------------------
| Method:
|	SetClientStrings
|
| Parameters:
|	das        - Data attributes.
|	clientName - The full service name.
|
| Description:
|	Replace the client name, client interface and realm strings
|	with those taken from the first principal found in the service
|	key table.
|
| Returns:
|	None.
|
| Raises:
|	None.
-----------------------------------------------------------------------------*/

static
void
SetClientStrings(
	Krb5TcpProxyDAs_t	*das,
	char				*clientName )
{
	auto	char	*end;

	/*	Discard full service name.	*/
	if ( das->fullClientName )
		free(
			das->fullClientName );
	das->fullClientName =
		strdup(
			clientName );

	/*	Discard any existing strings.	*/
	if ( das->clientName )
		free(
			das->clientName );

	/*	Create separate service name and interface strings.	*/
	if ( (end =
		strchr(
			clientName,
			'/' )) == 0)
	{
		/*	Raise exception.	*/
		rb_raise(
			eKrb5TcpProxy,
			"SetClientStrings (%s %d) - Missing interface\n"
			"   in principal (%s)",
			__FILE__,
			__LINE__,
			das->fullClientName );
	}

	/*	Terminate string and dulicate service name.	*/
	*end             = 0;
	das->clientName =
		strdup(
			clientName );
	strcpy(
		clientName,
		end + 1 );

	/*	Discard existing interface string.	*/
	if ( das->clientInterface )
		free(
			das->clientInterface );

	/*	Split remaining into interface and realm.	*/
	if ( (end =
		strchr(
			clientName,
			'@' )) == 0)
	{
		/*	Raise exception.	*/
		rb_raise(
			eKrb5TcpProxy,
			"SetClientStrings (%s %d) - Missing realm\n"
			"   in principal (%s)",
			__FILE__,
			__LINE__,
			das->fullClientName );
	}

	/*	Duplicate string to get realm.	*/
	*end                  = 0;
	das->clientInterface =
		strdup(
			clientName );
	strcpy(
		clientName,
		end + 1 );

	/*	Discard existing realm string.	*/
	if ( das->realm )
		free(
			das->realm );

	/*	Duplicate realm string.	*/
	das->realm =
		strdup(
			clientName );

	/*	Discard unparsed server principal name string.	*/
	free(
		clientName );
}

/*-----------------------------------------------------------------------------
| Method:
|	AuthenticateServiceToService
|
| Parameters:
|	serverContext - The server context.
|	das			  - The Kerberos 5 Client data attributes.
|
| Description:
|	Authenticate a service to another service.
|
| Returns:
|	None.
|
| Raises:
|	ExceptionKrb5TcpProxy on credential table initialization errors and
|	credential caching errors.
-----------------------------------------------------------------------------*/

static
void
AuthenticateServiceToService(
	Krb5TcpProxyDAs_t		*das,
	Krb5ServerContext_t		*serverContext )
{
	auto	krb5_get_init_creds_opt	options;
	auto	krb5_error_code			error;
	auto	krb5_creds				inCredentials;
	auto	krb5_creds				*outCredentials;
	auto	time_t					now;
	auto	time_t					endTime;

	/*	Delete credential contents.	*/
	if ( das->credentials.server )
	{
		/*	Don't discard client principal.	*/
		das->credentials.client = 0;

		/*	Do discard credentials.	*/
		krb5_free_cred_contents(
			serverContext->context,
			&das->credentials );
	}

	/*	Attempt to get credentials from cache.	*/
	memset(
		&inCredentials,
		0,
		sizeof( inCredentials ));
	inCredentials.client = das->clientPrincipal;
	inCredentials.server = das->tgtPrincipal;

	/*	Attempt to get the credentials.	*/
	if (
		krb5_get_credentials(
			serverContext->context,
			0,
			serverContext->defaultCache,
			&inCredentials,
			&outCredentials ) == 0)
	{
		/*	Get current time.	*/
		now     =
			time( 0 );
		endTime = outCredentials->times.endtime;

		/*	Return if lots of time left.	*/
		if (now + EXPIRE_TIME < endTime)
		{
			/*	Discard credentials.	*/
			krb5_free_creds(
				serverContext->context,
				outCredentials );
			return;
		}
	}

	/*	Either we don't have the TGT, or it is about to expire.
	*	Get credentials from key table.
	*/
	krb5_get_init_creds_opt_init(
		&options );
	krb5_get_init_creds_opt_set_tkt_life(
		&options,
		TICKET_LIFE );
	krb5_get_init_creds_opt_set_forwardable(
		&options,
		1 );
	if ( (error =
		krb5_get_init_creds_keytab(
			serverContext->context,
			&das->credentials,
			das->clientPrincipal,
			serverContext->keyTable,
			0,
			das->tgtServiceName,
			&options )) != 0)
	{
		/*	Raise exception.	*/
		Krb5_Error(
			error,
			"AuthenticateServiceToService" );
		rb_raise(
			eKrb5TcpProxy,
			"AuthenticateServiceToService (%s %d) - Can't get initial\n"
			"   credentials from keytab",
			__FILE__,
			__LINE__ );
	}

	/*	Store acquired credentials.	*/
	if ( (error =
		krb5_cc_store_cred(
			serverContext->context,
			serverContext->defaultCache,
			&das->credentials )) != 0)
	{
		/*	Raise exception.	*/
		Krb5_Error(
			error,
			"AuthenticateServiceToService: " );
		rb_raise(
			eKrb5TcpProxy,
			"AuthenticateServiceToService (%s %d) - Can't store\n"
			"   credentials in cache",
			__FILE__,
			__LINE__ );
	}
}

/*-----------------------------------------------------------------------------
| Method:
|	Free
|
| Parameters:
|	dataAttributes - The data attributes to free.
|
| Description:
|	Free the Kerberos 5 data associated with the object.
|
| Returns:
|	None.
|
| Raises:
|	None.
-----------------------------------------------------------------------------*/

static
void
Krb5TcpProxy_Free(
	void	*dataAttributes )
{
	auto	Krb5TcpProxyDAs_t		*das;
	auto	Krb5ServerContext_t		*sc;

	/*	Get Kerberos 5 data structure pointer.	*/
	das = (Krb5TcpProxyDAs_t *) dataAttributes;
	sc  = &das->serverContext;

	/*	Free strings.	*/
	SAFE_FREE( das->tgtServiceName   );
	SAFE_FREE( das->fullClientName   );
	SAFE_FREE( das->clientName       );
	SAFE_FREE( das->realm            );
	SAFE_FREE( das->serviceName      );
	SAFE_FREE( das->serviceInterface );
	SAFE_FREE( das->clientInterface  );

	/*	Discard Kerberos5 stuff.	*/
	if ( das->clientPrincipal )
		krb5_free_principal(
			sc->context,
			das->clientPrincipal );
	if ( das->tgtPrincipal )
		krb5_free_principal(
			sc->context,
			das->tgtPrincipal );
	if ( das->authorizationContext )
		krb5_auth_con_free(
			sc->context,
			das->authorizationContext );
	if ( das->credentials.server )
		krb5_free_cred_contents(
			sc->context,
			&das->credentials );

	/*	Discard Kerberos5 stuff.	*/
	if ( sc->serverPrincipal )
		krb5_free_principal(
			sc->context,
			sc->serverPrincipal );
	if ( sc->keyTable )
		krb5_kt_close(
			sc->context,
			sc->keyTable );
	if ( sc->context )
		krb5_free_context(
			sc->context );

	/*	Zero data attribute fields.	*/
	sc->keyTable        = 0;
	sc->defaultCache    = 0;
	sc->serverPrincipal = 0;
	sc->context         = 0;

	/*	If the socket is still open, close it.	*/
	if (das->sock >= 0)
	{
		/*	Closet the socket.	*/
		close(
			das->sock );
	}

	/*	Zero structure fields.	*/
	memset(
		das,
		0,
		sizeof( Krb5TcpProxyDAs_t ));
}

/*-----------------------------------------------------------------------------
| Method:
|	Allocate
|
| Parameters:
|	klass - The pointer to this Ruby object's class.
|
| Description:
|	Allocate a Krb5TcpProxy object.
|
| Returns:
|	Returns newly allocated object.
|
| Raises:
|	Nothing.
-----------------------------------------------------------------------------*/

static
VALUE
Krb5TcpProxy_Allocate(
	VALUE	klass )
{
	auto	VALUE				object;
	auto	Krb5TcpProxyDAs_t	*das;

	/*	Allocate kerberos 5 data structure.	*/
	object =
		Data_Make_Struct(
			klass,
			Krb5TcpProxyDAs_t,
			0,
			Krb5TcpProxy_Free,
			das );
	das->sock = -1;

	/*	Return new object.	*/
	return object;
}

/*-----------------------------------------------------------------------------
| Method:
|	Initialize
|
| Parameters:
|	self - The pointer to the this Ruby object.
|
| Description:
|	Initialize the Krb5TcpProxy object.
|
|	By convention, the service name should be the class name of the
|	ActiveObject sub-class that this is a listener for.
|
|	The service interface string should include the DNS name of the
|	interface |	of which the ServiceListener that is a sub-class of
|	this is listening on.
|
| Returns:
|	Returns super result.
|
| Raises:
|	ExceptionKrb5TcpProxy on out of memory, context initialization and
|	principal building errors.
-----------------------------------------------------------------------------*/

static
VALUE
Krb5TcpProxy_Initialize(
	VALUE	self,
	VALUE	rClientName )
{
	auto	int						error;
	auto	Krb5TcpProxyDAs_t		*das;
	auto	Krb5ServerContext_t		*sc;
	auto	char					*clientName;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		Krb5TcpProxyDAs_t,
		das );
	sc = &das->serverContext;

	/*	Get the kerberos 5 context.	*/
	if (
		krb5_init_context(
			&sc->context ) )
	{
		/*	Raise exception.	*/
		sc->context = 0;
		rb_raise(
			eKrb5TcpProxy,
			"Krb5TcpProxy_Initialize (%s %d) - Can't initialize\n"
			"   Kerberos 5 context",
			__FILE__,
			__LINE__ );
	}

	/*	Get first principal from key table and extract
	*	principal name strings into the data attributes.
	*/
	clientName =
		GetFirstPrincipalFromKeyTable(
			das,
			sc,
			RSTRING( rClientName )->ptr );

	/*	Set client name strings.	*/
	SetClientStrings(
		das,
		clientName );

	/*	Create ticket cache file.	*/
	CreateTicketCache(
		das );

	/*	Build TGT principal.	*/
	if ( (error =
		krb5_build_principal(
			sc->context,
			&das->tgtPrincipal,
			strlen(
				das->realm ),
			das->realm,
			"krbtgt",
			das->realm,
			0 )) != 0)
	{
		/*	Raise exception.	*/
		rb_raise(
			eKrb5TcpProxy,
			"Krb5TcpProxy_Initialize (%s %d) - Can't build\n"
			"   TGT principal (krbtgt/%s)",
			__FILE__,
			__LINE__,
			das->realm );
	}

	/*	Create TGT service name string.	*/
	if ( (das->tgtServiceName =
		calloc(
			1,
			strlen(
				das->realm ) * 2 + 10 )) != 0)
	{
		/*	Create name of TGT granting service.	*/
		sprintf(
			das->tgtServiceName,
			"krbtgt/%s@%s",
			das->realm,
			das->realm );
	}
	else
	{
		/*	Raise exception.	*/
		rb_raise(
			eKrb5TcpProxy,
			"Krb5TcpProxy_Initialize (%s %d) - Out of memory",
			__FILE__,
			__LINE__ );
	}

	/*	Invoke super class initialization.	*/
	return
		rb_call_super(
			0,
			0 );
}

/*-----------------------------------------------------------------------------
| Method:
|	Connect
|
| Parameters:
|	self         - The pointer to the this Ruby object.
|	rPort        - Port to connect to.
|	rInterface   - DNS name of interface of port to connect to.
|	rServiceName - Service name string.
|
| Description:
|	Connect this Krb5TcpProxy socket to a service.
|
| Returns:
|	Returns zero.
|
| Raises:
|	ExceptionKrb5TcpProxy on principal resolution and authentication errors.
-----------------------------------------------------------------------------*/

static
VALUE
Krb5TcpProxy_Connect(
	VALUE	self,
	VALUE	rPort,
	VALUE	rInterface,
	VALUE	rServiceName )
{
	auto	Krb5TcpProxyDAs_t	*das;
	auto	Krb5ServerContext_t	*sc;
	auto	long				length;
	auto	unsigned short		port;
	auto	int					error;
	auto	krb5_error			*errorReturn;
	auto	krb5_error_code		code;
	auto	VALUE				object;
	auto	krb5_data			data;
	auto	VALUE				values[ 3 ];
	auto	char				*string;
	auto	int					sizeOf;
	auto	int					localAddress;
	auto	struct sockaddr_in	clientAddress;
	auto	krb5_address		address;
	auto	krb5_rcache			replayCache;
	auto	krb5_data			replayCacheName;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		Krb5TcpProxyDAs_t,
		das );
	sc = &das->serverContext;

	/*	Initialize the service name and interface
	*	strings, and save the service port number.
	*/
	das->servicePort      =
		NUM2INT(
			rPort );
	das->serviceInterface =
		strdup(
			rb_str2cstr(
				rInterface,
				&length ));
	das->serviceName      =
		strdup(
			rb_str2cstr(
				rServiceName,
				&length ));

	/*	Resolve service name.	*/
	if (
		krb5_sname_to_principal(
			sc->context,
			das->serviceInterface,
			das->serviceName,
			KRB5_NT_SRV_HST,
			&sc->serverPrincipal ) ) 
	{
		/*	Raise exception.	*/
		rb_raise(
			eKrb5TcpProxy,
			"Krb5TcpProxy_Connect (%s:%d) - Bad Kerberos5\n"
			"   service principal (%s/%s)",
			__FILE__,
			__LINE__,
			das->serviceName,
			das->serviceInterface );
	}

	/*	Forward to super class.	*/
	values[ 0 ] = rPort;
	values[ 1 ] = rInterface;
	values[ 2 ] = rServiceName;
	object =
		rb_call_super(
			3,
			values );

	/*	Authenticate service to service.	*/
	AuthenticateServiceToService(
		das,
		&das->serverContext );

	/*	Send some data?	*/
    data.data   = das->serviceName;
    data.length = strlen( das->serviceName );

	/*	Send authentication.	*/
	if ( (code =
		krb5_sendauth(
			sc->context,
			&das->authorizationContext,
			(krb5_pointer) &das->sock,
			"1",
			das->clientPrincipal,
			sc->serverPrincipal,
			AP_OPTS_MUTUAL_REQUIRED,
			&data,
			NULL,
			sc->defaultCache,
			&errorReturn,
			NULL,
			NULL )) != 0)
	{
		/*	Communicate error.	*/
		com_err(
			das->serviceName,
			code,
			"\n   Return from krb5_sendauth().");

		/*	Discard Kerberos 5 stuff.	*/
		if ( errorReturn )
			krb5_free_error(
				sc->context,
				errorReturn );
		if ( sc->serverPrincipal )
			krb5_free_principal(
				sc->context,
				sc->serverPrincipal );

		/*	Zero all fields.	*/
		sc->serverPrincipal = 0;
		das->krb5AuthFlag   = 0;

		/*	Raise authentication exception.	*/
		rb_raise(
			eKrb5TcpProxy,
			"Krb5TcpProxy_Connect (%s:%d) - Failed authentication\n"
			"   to service (%s/%s)",
			__FILE__,
			__LINE__,
			das->serviceName,
			das->serviceInterface );
	}

	/*	Set the authorization flag.	*/
	das->krb5AuthFlag = 1;

	/*	Set remote addresses into authorization context.	*/
	SetLocalAddress(
		das->sock,
		das->authorizationContext,
		sc,
		das->serviceName,
		eKrb5TcpProxy );

	/*	Create replay cache.	*/
	CreateReplayCache(
		das->sock,
		das->authorizationContext,
		sc,
		eKrb5TcpProxy );

	/*	Return zero.	*/
	return
		INT2NUM(
			0 );
}

/*-----------------------------------------------------------------------------
| Method:
|	Close
|
| Parameters:
|	self - The pointer to the this Ruby object.
|
| Description:
|	Close the listening socket.  Allow the free routine to clean up the
|	Kerberos 5 data.
|
| Returns:
|	Returns nil.
|
| Raises:
|	Nothing.
-----------------------------------------------------------------------------*/

static
VALUE
Krb5TcpProxy_Close(
	VALUE	self)
{
	auto	Krb5TcpProxyDAs_t	*das;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		Krb5TcpProxyDAs_t,
		das );

	/*	Close socket.	*/
	if (das->sock >= 0)
	{
		/*	Close the socket.	*/
		close(
			das->sock );
		das->sock = -1;
	}

	/*	Return nil.	*/
	return Qnil;
}

static
void
Initialize_Krb5TcpProxy(
	void )
{
	/*	Create Krb5TcpProxy class.	*/
	cKrb5TcpProxy =
		rb_define_class(
			"Krb5TcpProxy",
			cKrb5TcpSocket );

	/*	Add allocation method.	*/
	rb_define_alloc_func(
		cKrb5TcpProxy,
		Krb5TcpProxy_Allocate );

	/*	Define methods in class.	*/
	rb_define_method(
		cKrb5TcpProxy,
		"initialize",
		Krb5TcpProxy_Initialize,
		1 );
	rb_define_method(
		cKrb5TcpProxy,
		"connect",
		Krb5TcpProxy_Connect,
		3 );
	rb_define_method(
		cKrb5TcpProxy,
		"close",
		Krb5TcpProxy_Close,
		0 );

	/*	Create exception class.	*/
	eKrb5TcpProxy =
		rb_define_class(
			"ExceptionKrb5TcpProxy",
			rb_eStandardError );
}

void
Init_RubyKerberos(
	void )
{
	Initialize_Krb5TcpSocket();
	Initialize_Krb5TcpServer();
	Initialize_Krb5TcpClient();
	Initialize_Krb5TcpConnection();
	Initialize_Krb5TcpProxy();
}
