/******************************************************************************
* Module:
*	TcpSocket
*
* Description:
*	A TCP Socket base class for Ruby.  This was done because sub-classing
*	the built-ins turned out to be nigh on impossible.
*
* Author:
*	John W. M. Stevens
******************************************************************************/

#include	<stdio.h>
#include	<stdlib.h>
#include	<errno.h>
#include	<limits.h>
#include	<string.h>
#include	<stdint.h>

#include	"ruby.h"
#include	"intern.h"

#include	"TcpSocket.h"

/*	TCP Socket Object.	*/
		VALUE	cTcpSocket;
static	VALUE	eTcpSocket;

/*-----------------------------------------------------------------------------
| Method:
|	Free
|
| Parameters:
|	dataAttributes - The data attributes to free.
|
| Description:
|	Free the data attribute structure of a TCP socket object, and close
|	the OS socket.
|
| Returns:
|	None.
|
| Raises:
|	None.
-----------------------------------------------------------------------------*/

static
void
TcpSocket_Free(
	void	*dataAttributes )
{
	auto	TcpSocketDAs_t	*das;

	/*	Get Kerberos 5 data structure pointer.	*/
	das = (TcpSocketDAs_t *) dataAttributes;

	/*	Discard strings.	*/
	if ( das->clientInterface )
		free(
			das->clientInterface );
	if ( das->serviceInterface )
		free(
			das->serviceInterface );
	if ( das->serviceName )
		free(
			das->serviceName );

	/*	Is the socket open?	*/
	if (das->sock >= 0)
	{
		/*	Yes.  Close it.	*/
		close(
			das->sock );
	}

	/*	Zero structure fields.	*/
	memset(
		das,
		0,
		sizeof( TcpSocketDAs_t ));
}

/*-----------------------------------------------------------------------------
| Method:
|	Allocate
|
| Parameters:
|	klass - The pointer to this Ruby object's class.
|
| Description:
|	Allocate the data attribute structure of the TcpSocket object.
|
| Returns:
|	Returns newly allocated object.
|
| Raises:
|	None.
-----------------------------------------------------------------------------*/

static
VALUE
TcpSocket_Allocate(
	VALUE	klass )
{
	auto	VALUE			object;
	auto	TcpSocketDAs_t	*das;

	/*	Allocate TCP socket data attributes structure.	*/
	object =
		Data_Make_Struct(
			klass,
			TcpSocketDAs_t,
			0,
			TcpSocket_Free,
			das );
	das->sock = -1;

	/*	Return new object.	*/
	return object;
}

/*-----------------------------------------------------------------------------
| Method:
|	Initialize
|
| Parameters:
|	argc - The number of method parameters.
|	argv - The method parameter list.
|	self - The pointer to the this Ruby object.
|
| Description:
|	Initialize the TcpSocket object.  With no parameters (arity zero),
|	this creates a new socket that will return a TcpSocket object on
|	accept.  With two parameters, an existing socket file descriptor
|	will be wrapped into a TcpSocket object.  With one parameter, a
|	new TcpSocket object will be created that on accept returns a new
|	instance of the passed in class.
|
| Returns:
|	Returns self.
|
| Raises:
|	ExceptionTcpSocket.
-----------------------------------------------------------------------------*/

static
VALUE
TcpSocket_Initialize(
	int		argc,
	VALUE	*argv,
	VALUE	self )
{
	auto	TcpSocketDAs_t		*das;
	struct	protoent			*protocolType;
	auto	int					protocol;
	auto	int					true;
	auto	char				*address;
	auto	long				length;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		TcpSocketDAs_t,
		das );

	/*	Is this a wrap of an existing socket?	*/
	if (argc == 2)
	{
		/*	Yes, it is.  Initialize the socket file descriptor
		*	data attribute, and the local address.
		*/
		das->sock =
			NUM2INT(
				argv[ 0 ] );

		/*	Is the address of the right length?	*/
		if (RSTRING( argv[ 1 ] )->len == sizeof( das->serviceAddress ))
			/*	Yes.  Save the internet address structure.	*/
			memcpy(
				&das->serviceAddress,
				RSTRING( argv[ 1 ] )->ptr,
				sizeof( das->serviceAddress ));
		else
			rb_raise(
				eTcpSocket,
				"TcpSocket_Initialize (%s:%d) - Address\n"
				"   wrong length (%d)",
				__FILE__,
				__LINE__,
				RSTRING( argv[ 1 ] )->len );

		/*	Return self.	*/
		return self;
	}
	else if (argc == 1)
	{
		/*	No, It's not.  Save the class
		*	of object to create on accept.
		*/
		das->klass = argv[ 0 ];
	}
	else
		/*	Save self class as class of object to create on accept.	*/
		das->klass = cTcpSocket;

	/*	Get the protocol number for tcp.	*/
	if ( (protocolType =
		getprotobyname(
			"tcp" )) == NULL)
		protocol = 6;
	else
		protocol = protocolType->p_proto;

	/*	Can we create the socket?	*/
	if ( (das->sock =
		socket(
			AF_INET,
			SOCK_STREAM,
			protocol )) == -1)
	{
		/*	No.  Throw socket creation error.	*/
		rb_raise(
			eTcpSocket,
			"TcpSocket_Initialize (%s:%d) - Socket creation error,\n"
			"   %s",
			__FILE__,
			__LINE__,
			strerror(
				errno ) );
	}

	/*	Can we turn off the NAGLE algorithm?	*/
	true = 1;
	if (
		setsockopt(
			das->sock,
			IPPROTO_TCP,
			TCP_NODELAY,
			&true,
			sizeof( true )) == -1)
	{
		/*	No.  Throw socket option exception.	*/
		rb_raise(
			eTcpSocket,
			"TcpSocket_Initialize (%s:%d) - Disable NAGEL option\n"
			"   setting error, %s",
			__FILE__,
			__LINE__,
			strerror(
				errno ) );
	}

	/*	Can we set the socket to be quickly reused?	*/
	true = 1;
	if (
		setsockopt(
			das->sock,
			SOL_SOCKET,
			SO_REUSEADDR,
			&true,
			sizeof( true )) == -1)
	{
		/*	No.  Throw socket option exception.	*/
		rb_raise(
			eTcpSocket,
			"TcpSocket_Initialize (%s:%d) - Reuse address option\n"
			"   setting error, %s",
			__FILE__,
			__LINE__,
			strerror(
				errno ) );
	}

	/*	Return self.	*/
	return self;
}

/*-----------------------------------------------------------------------------
| Method:
|	Connect
|
| Parameters:
|	self         - The pointer to the this Ruby object.
|	rPort        - Port to connect to.
|	rInterface   - DNS name of interface of port to connect to.
|	rServiceName - Port to connect to.
|
| Description:
|	Connect this TcpSocket socket to a service.
|
| Returns:
|	Returns zero.
|
| Raises:
|	DNS resolution and connect errors.
-----------------------------------------------------------------------------*/

static
VALUE
TcpSocket_Connect(
	VALUE	self,
	VALUE	rPort,
	VALUE	rInterface,
	VALUE	rServiceName )
{
	auto	TcpSocketDAs_t		*das;
	auto	int					error;
	auto	unsigned short		port;
	auto	long				address;
	auto	long				length;
	struct	sockaddr_in			socketAddress;
	auto	uint8_t				*buffer;
	auto	struct hostent		*hostEntry;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		TcpSocketDAs_t,
		das );

	/*	Get interface, service and port number.	*/
	das->servicePort =
		NUM2INT(
			rPort );
	das->serviceInterface =
		strdup(
			rb_str2cstr(
				rInterface,
				&length ));
	das->serviceName =
		strdup(
			rb_str2cstr(
				rServiceName,
				&length ));

	/*	Can we resolve interface DNS name to address?	*/
	if ( (hostEntry =
		gethostbyname(
			das->serviceInterface )) == NULL)
	{
		/*	No.  Raise DNS name resolution error.	*/
		rb_raise(
			eTcpSocket,
			"TcpSocket_Connect (%s:%d) - DNS resolution error\n"
			"   for (%s), %s",
			__FILE__,
			__LINE__,
			das->clientInterface,
			strerror(
				errno ) );
	}

	/*	Get address and port number.	*/
	address = *((long *) hostEntry->h_addr_list[ 0 ]);
	port    =
		htons(
			(unsigned short)
			das->servicePort );

	/*	Set up an AF_INET style address.	*/
	das->serviceAddress.sin_family      = AF_INET;
	das->serviceAddress.sin_port        = port;
	das->serviceAddress.sin_addr.s_addr = address;

	/*	Can we connect to the service?	*/
	if ( (error =
		connect(
			das->sock,
			(struct sockaddr *) &das->serviceAddress,
			sizeof( das->serviceAddress ) )) == -1)
	{
		/*	No.  Throw socket connection error.	*/
		rb_raise(
			eTcpSocket,
			"TcpSocket_Connect (%s:%d) - Socket connect error\n"
			"   to (%s at %d), %s",
			__FILE__,
			__LINE__,
			das->serviceInterface,
			das->servicePort,
			strerror(
				errno ) );
	}

	/*	Return zero.	*/
	return
		INT2NUM(
			0 );
}

/*-----------------------------------------------------------------------------
| Method:
|	WriteMessage
|
| Parameters:
|	argc - Number of parameters.
|	argv - Vector of parameters.
|	self - The pointer to the this Ruby object.
|
| Description:
|	Write a message to the socket.  A message is a four byte length followed
|	by that number of bytes.
|
| Returns:
|	Returns number of bytes written to socket minus the message length.
|
| Raises:
|	ExceptionTcpSocket on write errors.
-----------------------------------------------------------------------------*/

static
VALUE
TcpSocket_WriteMessage(
	int		argc,
	VALUE	*argv,
	VALUE	self )
{
	register	int				i;
	auto		uint8_t			*buffer;
	auto		long			length;
	auto		long			sum;
	auto		long			total;
	auto		TcpSocketDAs_t	*das;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		TcpSocketDAs_t,
		das );

	/*	Get number of characters to write.	*/
	for (i = 0, sum = 0; i < argc; i++)
	{
		/*	Get buffers to write.	*/
		buffer =
			rb_str2cstr(
				argv[ i ],
				&length );

		/*	Sum amounts written.	*/
		sum += length;
	}

	/*	Write message length to socket.	*/
	total = sum;
	sum   =
		htonl(
			sum );
	write(
		das->sock,
		&sum,
		sizeof( sum ) );

	/*	Write buffers to socket.	*/
	for (i = 0, sum = 0; i < argc; i++)
	{
		/*	Get buffer.	*/
		buffer =
			rb_str2cstr(
				argv[ i ],
				&length );

		/*	Write data buffer to socket.	*/
		if (
			write(
				das->sock,
				buffer,
				length ) < 0)
		{
			/*	Write error.	*/
			rb_raise(
				eTcpSocket,
				"TcpSocket_WriteMessage (%s:%d) - Socket write error,\n"
				"   %s",
				__FILE__,
				__LINE__,
				strerror(
					errno ) );
		}
	}

	/*	Return number of bytes written to socket.	*/
	return
		INT2NUM(
			total );
}

/*-----------------------------------------------------------------------------
| Method:
|	ReadMessage
|
| Parameters:
|	self - The pointer to the this Ruby object.
|
| Description:
|	Read a message from the socket.  Read the length of the message first,
|	convert length to machine coded integer, then read that number of bytes
|	from the socket in blocking mode.
|
| Returns:
|	Returns message string.
|
| Raises:
|	ExceptionTcpSocket on read errors.
-----------------------------------------------------------------------------*/

static
VALUE
TcpSocket_ReadMessage(
	VALUE	self )
{
	register	int				i;
	auto		int				size;
	auto		VALUE			message;
	auto		uint8_t			*buffer;
	auto		long			length;
	auto		long			sum;
	auto		long			total;
	auto		TcpSocketDAs_t	*das;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		TcpSocketDAs_t,
		das );

	/*	Can we read the message length?	*/
	if ( (size =
		read(
			das->sock,
			&sum,
			sizeof( sum ) )) < 0)
	{
		/*	No.  Raise exception.	*/
		rb_raise(
			eTcpSocket,
			"TcpSocket_ReadMessage (%s:%d) - Message length read error,\n"
			"   %s",
			__FILE__,
			__LINE__,
			strerror(
				errno ) );
	}
	else if (size == 0)
	{
		/*	End of file read error.	*/
		rb_raise(
			eTcpSocket,
			"TcpSocket_ReadMessage (%s:%d) - No message to read,\n"
			"   %s",
			__FILE__,
			__LINE__,
			strerror(
				errno ) );
	}

	/*	Convert from network order to host order.	*/
	sum =
		ntohl(
			sum );

	/*	Allocate a buffer.	*/
	if ( (buffer =
		malloc(
			sum + 2 )) == 0)
	{
		/*	Memory allocation error.	*/
		rb_raise(
			eTcpSocket,
			"TcpSocket_ReadMessage (%s:%d) - Out of memory"
			__FILE__,
			__LINE__ );
	}

	/*	Can we read the message?	*/
	if ( (total =
		read(
			das->sock,
			buffer,
			sum )) < 0)
	{
		/*	No.  Free buffer and raise read error.	*/
		free(
			buffer );
		rb_raise(
			eTcpSocket,
			"TcpSocket_ReadMessage (%s:%d) - Socket read error,\n"
			"   %s",
			__FILE__,
			__LINE__,
			strerror(
				errno ) );
	}

	/*	Did we read as many bytes as we needed to?	*/
	if (total != sum)
	{
		/*	No.  Raise read exception.	*/
		free(
			buffer );
		rb_raise(
			eTcpSocket,
			"TcpSocket_ReadMessage (%s:%d) - Short read error,\n"
			"   sum = %ld, but total = %ld",
			__FILE__,
			__LINE__,
			sum,
			total );
	}

	/*	Return message.	*/
	message =
		rb_str_new(
			buffer,
			sum );
	free(
		buffer );
	return message;
}

/*-----------------------------------------------------------------------------
| Method:
|	Close
|
| Parameters:
|	self - The pointer to the this Ruby object.
|
| Description:
|	Close the socket.
|
| Returns:
|	Returns nil.
|
| Raises:
|	Nothing.
-----------------------------------------------------------------------------*/

static
VALUE
TcpSocket_Close(
	VALUE	self)
{
	auto	TcpSocketDAs_t	*das;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		TcpSocketDAs_t,
		das );

	/*	Close socket.	*/
	if (das->sock >= 0)
	{
		/*	Close the socket.	*/
		close(
			das->sock );
		das->sock = -1;
	}

	/*	Return nil.	*/
	return Qnil;
}

/*-----------------------------------------------------------------------------
| Method:
|	FileNo
|
| Parameters:
|	self - The pointer to the this Ruby object.
|
| Description:
|	Return the socket file descriptor number.
|
| Returns:
|	Returns the file number
|
| Raises:
|	Nothing.
-----------------------------------------------------------------------------*/

static
VALUE
TcpSocket_FileNo(
	VALUE	self)
{
	auto	TcpSocketDAs_t		*das;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		TcpSocketDAs_t,
		das );

	/*	Return file number.	*/
	return
		INT2NUM(
			das->sock );
}

/*-----------------------------------------------------------------------------
| Method:
|	ListenOn
|
| Parameters:
|	self         - The pointer to the this Ruby object.
|	rPort        - Port number to listen on.
|	rInterface   - If present, specifies the local interface to bind to.
|                  if missing, the port will be bound into all local
|                  interfaces.
|	rServiceName - Service name to listen for.
|
| Description:
|	Bind to a given port and address and start listening for connections.
|
| Returns:
|	Returns nil.
|
| Raises:
|	ExceptionTcpSocket for bind, listen and DNS name resolution errors.
-----------------------------------------------------------------------------*/

static
VALUE
TcpSocket_ListenOn(
	VALUE	self,
	VALUE	rPort,
	VALUE	rInterface,
	VALUE	rServiceName )
{
	auto	int					port;
	auto	int					on;
	auto	TcpSocketDAs_t		*das;
	auto	long				address;
	auto	long				length;
	auto	struct hostent		*hostEntry;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		TcpSocketDAs_t,
		das );

	/*	Extract service port.	*/
	das->servicePort =
		NUM2INT(
			rPort );

	/*	Is the interface string empty?	*/
	if (rInterface == Qnil || RSTRING( rInterface )->len == 0)
	{
		/*	Set address to any.	*/
		address = INADDR_ANY;

		/*	Set service interface to empty.	*/
		das->serviceInterface = 0;
	}
	else
	{
		/*	Get local interface DNS name.	*/
		das->serviceInterface =
			strdup(
				rb_str2cstr(
					rInterface,
					&length ));

		/*	Can we resolve interface DNS name to address?	*/
		if ( (hostEntry =
			gethostbyname(
				das->serviceInterface )) == NULL)
		{
			/*	Resolve DNS name to address.	*/
			rb_raise(
				eTcpSocket,
				"TcpSocket_ListenOn (%s:%d) - DNS resolution error\n"
				"   for (%s), %s",
				__FILE__,
				__LINE__,
				das->serviceInterface,
				strerror(
					errno ) );
		}

		/*	Get address and port number.	*/
		address = *((long *) hostEntry->h_addr_list[ 0 ]);
	}

	/*	Get service name.	*/
	das->serviceName =
		strdup(
			rb_str2cstr(
				rServiceName,
				&length ));

	/*	Create local address.	*/
	das->serviceAddress.sin_family      = AF_INET;
	das->serviceAddress.sin_addr.s_addr = address;
	das->serviceAddress.sin_port        =
		htons(
			(unsigned short) das->servicePort );

	/*	Bind to address.	*/
	if (
		bind(
			das->sock,
			(struct sockaddr *) &das->serviceAddress,
			sizeof( das->serviceAddress )) == -1)
	{
		/*	Report bind error.	*/
		rb_raise(
			eTcpSocket,
			"TcpSocket_ListenOn (%s:%d) - Bind error\n"
			"   for (%s), %s",
			__FILE__,
			__LINE__,
			das->serviceInterface,
			strerror(
				errno ) );
	}

	/*	Set resuse immediately flag.	*/
	on = 1;
	if (
		setsockopt(
			das->sock,
			SOL_SOCKET,
			SO_REUSEADDR,
			(char *) &on,
			sizeof( on )) )
	{
		/*	Raise exception on accept error.	*/
		rb_raise(
			eTcpSocket,
			"TcpSocket_Accept (%s:%d) - Socket set error,\n"
			"   %s",
			__FILE__,
			__LINE__,
			strerror(
				errno ) );
	}

	/*	Listen for connections.	*/
	if (
		listen(
			das->sock,
			5 ) == -1)
	{
		/*	Report bind error.	*/
		rb_raise(
			eTcpSocket,
			"TcpSocket_ListenOn (%s:%d) - Listen error\n"
			"   for (%s), %s",
			__FILE__,
			__LINE__,
			das->serviceInterface,
			strerror(
				errno ) );
	}

	/*	Return nil.	*/
	return Qnil;
}

/*-----------------------------------------------------------------------------
| Method:
|	Accept
|
| Parameters:
|	argc - The number of method parameters.
|	argv - The method parameter list.
|	self - The pointer to the this Ruby object.
|
| Description:
|	Accept connections, create new objects that implement an initializer
|	that takes a socket descriptor number and an internet address
|	structure packed into a string.
|
| Returns:
|	Returns the new object created to wrap the accept results.  The
|	new object's class must be TcpSocket, or a sub-class thereof.
|
| Raises:
|	ExceptionTcpSocket for accept errors.
-----------------------------------------------------------------------------*/

static
VALUE
TcpSocket_Accept(
	int		argc,
	VALUE	*argv,
	VALUE	self )
{
	register	int					i;
	register	int					j;
	auto		VALUE				rNewSocket;
	auto		VALUE				values[ 8 ];
	auto		int					sock;
	auto		int					on;
	auto		TcpSocketDAs_t		*das;
	auto		socklen_t			addressLength;
	auto		VALUE				klass;

	/*	More than 6?	*/
	if (argc > 6)
	{
		/*	Raise exception on overflow.	*/
		rb_raise(
			eTcpSocket,
			"TcpSocket_Accept (%s:%d) - Accept parameter count error\n"
			"   %d, expected no more than 6",
			__FILE__,
			__LINE__,
			argc );
	}

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		TcpSocketDAs_t,
		das );

	/*	Accept new connection.	*/
	addressLength = sizeof( das->clientAddress );
	if ( (sock =
		accept(
			das->sock,
			(struct sockaddr *) &das->clientAddress,
			&addressLength )) < 0)
	{
		/*	Raise exception on accept error.	*/
		rb_raise(
			eTcpSocket,
			"TcpSocket_Accept (%s:%d) - Socket accept error,\n"
			"   %s",
			__FILE__,
			__LINE__,
			strerror(
				errno ) );
	}

	/*	Set resuse immediately flag.	*/
	on = 1;
	if (
		setsockopt(
			sock,
			SOL_SOCKET,
			SO_REUSEADDR,
			(char *) &on,
			sizeof( on )) )
	{
		/*	Raise exception on accept error.	*/
		rb_raise(
			eTcpSocket,
			"TcpSocket_Accept (%s:%d) - Socket set error,\n"
			"   %s",
			__FILE__,
			__LINE__,
			strerror(
				errno ) );
	}

	/*	Save socket descriptor number and client address.	*/
	i             = 0;
	values[ i++ ] =
		INT2NUM(
			sock );
	values[ i++ ] =
		rb_str_new(
			(char *) &das->clientAddress,
			addressLength );

	/*	Was there more than one parameter?	*/
	if (argc > 0)
	{
		/*	Yes.  Fill in rest of parameters.	*/
		for (j = 0; j < argc && j < 8; i++, j++)
			values[ i ] = argv[ j ];
	}

	/*	Create new instance of self, or of the class passed
	*	in at initialization time.
	*/
	rNewSocket =
		rb_class_new_instance(
			i,
			values,
			das->klass );

	/*	Return new socket object.	*/
	return rNewSocket;
}

/*-----------------------------------------------------------------------------
| Method:
|	Select
|
| Parameters:
|	argc - The number of method parameters.
|	argv - The method parameter list.
|	self - The pointer to the this Ruby object.
|
| Description:
|	Select on objects that implement the fileno interface.
|
| Returns:
|	Returns array of objects that satisfy the select criteria.
|
| Raises:
|	Accept errors.
-----------------------------------------------------------------------------*/

static
VALUE
TcpSocket_Select(
	int		argc,
	VALUE	*argv,
	VALUE	self )
{
	register	int				i;
	register	int				j;
	auto		fd_set			sets[ 3 ];
	auto		VALUE			object;
	auto		VALUE			ary;
	auto		VALUE			returnArray;
	auto		VALUE			setArray;
	auto		VALUE			fileDescriptor;
	auto		ID				symbol;
	auto		int				bits;
	auto		int				maxDescriptors;
	auto		int				descriptor;
	auto		struct timeval	delay;
	auto		struct timeval	*timeout;

	/*	Intern the symbol.	*/
	symbol =
		rb_intern(
			"fileno" );

	/*	For the number of parameters, do:	*/
	for (i              = 0,
		 maxDescriptors = 0;
		 
		 i < argc && i < 3;
		 
		 i++ )
	{
		/*	Zero out the file descriptor set.	*/
		FD_ZERO(
			sets + i );

		/*	Array.	*/
		if ((ary = argv[ i ]) == Qnil)
			/*	Skip the nils.	*/
			continue;

		/*	For the number of read elements.	*/
		for (j = 0; j < RARRAY( ary )->len; j++)
		{
			/*	File descriptor.	*/
			fileDescriptor =
				rb_funcall(
					RARRAY( ary )->ptr[ j ],
					symbol,
					0 );
			descriptor     =
				NUM2INT(
					fileDescriptor );

			/*	Determine maximum number of file descriptors.	*/
			maxDescriptors =
				(maxDescriptors < descriptor)
					? descriptor
					: maxDescriptors;

			/*	Set file descriptor in set.	*/
			FD_SET(
				descriptor,
				sets + i );
		}
	}

	/*	Do we have a time delay?	*/
	if (argc == 4)
	{
		/*	Yes.  Extract it.	*/
		delay.tv_usec = 0;
		delay.tv_sec  =
			NUM2INT(
				argv[ 3 ] );
		timeout = &delay;
	}
	else
		/*	No timeout.	*/
		timeout = 0;

	/*	Create array.	*/
	if ( (bits =
		select(
			maxDescriptors + 1,
			sets + 0,
			sets + 1,
			sets + 2,
			timeout )) == -1)
	{
		/*	Raise exception.	*/
		rb_raise(
			eTcpSocket,
			"TcpSocket_Select (%s:%d) - Socket select error,\n"
			"   %s",
			__FILE__,
			__LINE__,
			strerror(
				errno ) );
	}

	/*	Create new array.	*/
	returnArray =
		rb_ary_new2(
			3 );

	/*	Load the array with arrays of selected objects.	*/
	for (i = 0; i < argc && i < 3; i++)
	{
		/*	Array.	*/
		if ((ary = argv[ i ]) == Qnil)
		{
			/*	Push a nil back.	*/
			rb_ary_push(
				returnArray,
				Qnil );
			continue;
		}

		/*	Create ouput array.	*/
		setArray =
			rb_ary_new();

		/*	For the number of objects in the input array, do:	*/
		for (j   = 0,
			 ary = argv[ i ];

			 j < RARRAY( ary )->len;
			 
			 j++ )
		{
			/*	File descriptor.	*/
			object         = RARRAY( ary )->ptr[ j ];
			fileDescriptor =
				rb_funcall(
					object,
					symbol,
					0 );
			descriptor     =
				NUM2INT(
					fileDescriptor );

			/*	Is this fileDescriptor set?	*/
			if (
				FD_ISSET(
					descriptor,
					sets + i ) )
			{
				/*	Yes.  Append object into array.	*/
				rb_ary_push(
					setArray,
					object );
			}
		}

		/*	Push array to end of return array.	*/
		rb_ary_push(
			returnArray,
			setArray );
	}

	/*	Return new array of objects.	*/
	return returnArray;
}

/*-----------------------------------------------------------------------------
| Method:
|	Authenticate
|
| Parameters:
|	self - The pointer to the this Ruby object.
|
| Description:
|	Authenticate client to server.
|
| Returns:
|	Returns self.
|
| Raises:
|	Authentication failure.
-----------------------------------------------------------------------------*/

static
VALUE
TcpSocket_Authenticate(
	VALUE	self)
{
	/*	Return self.	*/
	return self;
}

void
Init_TcpSocket(
	void )
{
	/*	Create TcpSocket class.	*/
	cTcpSocket =
		rb_define_class(
			"TcpSocket",
			rb_cObject );

	/*	Add allocation method.	*/
	rb_define_alloc_func(
		cTcpSocket,
		TcpSocket_Allocate );

	/*	Define methods in class.	*/
	rb_define_method(
		cTcpSocket,
		"initialize",
		TcpSocket_Initialize,
		-1 );
	rb_define_method(
		cTcpSocket,
		"connect",
		TcpSocket_Connect,
		3 );
	rb_define_method(
		cTcpSocket,
		"close",
		TcpSocket_Close,
		0 );
	rb_define_method(
		cTcpSocket,
		"fileno",
		TcpSocket_FileNo,
		0 );
	rb_define_method(
		cTcpSocket,
		"writeMessage",
		TcpSocket_WriteMessage,
		-1 );
	rb_define_method(
		cTcpSocket,
		"readMessage",
		TcpSocket_ReadMessage,
		0 );
	rb_define_method(
		cTcpSocket,
		"listenOn",
		TcpSocket_ListenOn,
		3 );
	rb_define_method(
		cTcpSocket,
		"accept",
		TcpSocket_Accept,
		-1 );
	rb_define_method(
		cTcpSocket,
		"authenticate",
		TcpSocket_Authenticate,
		0 );
	rb_define_singleton_method(
		cTcpSocket,
		"select",
		TcpSocket_Select,
		-1 );

	/*	Create exception class.	*/
	eTcpSocket =
		rb_define_class(
			"ExceptionTcpSocket",
			rb_eStandardError );
}
