/******************************************************************************
* Module:
*	UdpSocket
*
* Description:
*	A UDP Socket base class.
*
* Author:
*	John W. M. Stevens
******************************************************************************/

#include	<stdio.h>
#include	<stdlib.h>
#include	<errno.h>
#include	<limits.h>
#include	<string.h>
#include	<stdint.h>
#include	<netdb.h>

#include	<sys/socket.h>
#include	<netinet/in.h>
#include	<netinet/tcp.h>

#include	"ruby.h"
#include	"intern.h"


/*	Kerberos server information.	*/
		VALUE	cUdpSocket;
static	VALUE	eUdpSocket;

#define	UdpSocket_DAS							\
	/*	Service side data.	*/					\
	char				*serviceName;			\
	char				*serviceInterface;		\
	int					servicePort;			\
	struct sockaddr_in	serviceAddress;			\
												\
	/*	Client side data.	*/					\
	char				*clientInterface;		\
	int					clientPort;				\
	struct sockaddr_in	clientAddress;			\
												\
	/*	Socket number and accept klass.	*/		\
	int					sock;					\
	VALUE				klass;

/*	Socket information.	*/
typedef	struct	UdpSocketDas_st
{
	UdpSocket_DAS
} UdpSocketDAs_t;

/*-----------------------------------------------------------------------------
| Method:
|	Free
|
| Parameters:
|	dataAttributes - The data attributes to free.
|
| Description:
|	Free the data attributes associated with the object.
|
| Returns:
|	None.
|
| Raises:
|	None.
-----------------------------------------------------------------------------*/

static
void
UdpSocket_Free(
	void	*dataAttributes )
{
	auto	UdpSocketDAs_t	*das;

	/*	Get Kerberos 5 data structure pointer.	*/
	das = (UdpSocketDAs_t	*) dataAttributes;

	/*	Discard strings.	*/
	if ( das->clientInterface )
		free(
			das->clientInterface );
	if ( das->serviceInterface )
		free(
			das->serviceInterface );
	if ( das->serviceName )
		free(
			das->serviceName );

	/*	Is the socket open?	*/
	if (das->sock >= 0)
		/*	Yes.  Close it.	*/
		close(
			das->sock );

	/*	Zero structure fields.	*/
	memset(
		das,
		0,
		sizeof( UdpSocketDAs_t ));
}

/*-----------------------------------------------------------------------------
| Method:
|	Allocate
|
| Parameters:
|	klass - The pointer to this Ruby object's class.
|
| Description:
|	Allocate a UdpSocket object.
|
| Returns:
|	Returns newly allocated object.
|
| Raises:
|	Nothing.
-----------------------------------------------------------------------------*/

static
VALUE
UdpSocket_Allocate(
	VALUE	klass )
{
	auto	VALUE			object;
	auto	UdpSocketDAs_t	*das;

	/*	Allocate kerberos 5 data structure.	*/
	object =
		Data_Make_Struct(
			klass,
			UdpSocketDAs_t,
			0,
			UdpSocket_Free,
			das );
	das->sock = -1;

	/*	Return new object.	*/
	return object;
}

/*-----------------------------------------------------------------------------
| Method:
|	Initialize
|
| Parameters:
|	self           - The pointer to the this Ruby object.
|	rPort          - The port number.
|	rSocketAddress - The socket address.
|
| Description:
|	Initialize the UdpSocket object with its address and port and set to
|	broadcast mode.
|
| Returns:
|	Returns newly initialized object.
|
| Raises:
|	ExceptionUdpSocket on create, bind, option setting and DNS
|	resolution errors.
-----------------------------------------------------------------------------*/

static
VALUE
UdpSocket_Initialize(
	VALUE	self,
	VALUE	rPort,
	VALUE	rAddress )
{
	auto	UdpSocketDAs_t		*das;
	auto	struct protoent		*protocolType;
	auto	int					protocol;
	auto	int					true;
	auto	long				length;
	auto	char				*addressString;
	auto	long				address;
	auto	unsigned short		netPort;
	auto	struct sockaddr_in	internetAddress;
	auto	struct hostent		*hostEntry;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		UdpSocketDAs_t,
		das );

	/*	Get address to write to.	*/
	addressString =
		rb_str2cstr(
			rAddress,
			&length );
	das->port     =
		INT2NUM(
			rPort );
	if (length > 0)
		das->interface =
			strdup(
				addressString );

	/*	Can we resolve interface DNS name to address?	*/
	if ( (hostEntry =
		gethostbyname(
			addressString )) == NULL)
	{
		/*	No.  Raise DNS name to address resolution error.	*/
		rb_raise(
			eUdpSocket,
			"UdpSocket_Initialize (%s:%d) - DNS resolution error\n"
			"   for (%s), %s\n",
			__FILE__,
			__LINE__,
			addressString,
			strerror(
				errno ) );
	}

	/*	Get address and port number.	*/
	address = *((long *) hostEntry->h_addr_list[ 0 ]);
	netPort =
		htons(
			(unsigned short)
			das->port );

	/*	Set up an AF_INET style address.	*/
	das->address.sin_family      = AF_INET;
	das->address.sin_port        = netPort;
	das->address.sin_addr.s_addr = address;

	/*	Get the protocol number for udp.	*/
	if ( (protocolType =
		getprotobyname(
			"udp" )) == NULL)
		protocol = 17;
	else
		protocol = protocolType->p_proto;

	/*	Can we create the UDP socket?	*/
	if ( (das->sock =
		socket(
			AF_INET,
			SOCK_DGRAM,
			protocol )) == -1)
	{
		/*	No.  Throw socket creation error.	*/
		rb_raise(
			eUdpSocket,
			"UdpSocket_Initialize (%s:%d) - UDP socket creation error,\n"
			"   %s\n",
			__FILE__,
			__LINE__,
			strerror(
				errno ) );
	}

	/*	Can we set the socket to broadcast mode?	*/
	true = 1;
	if (
		setsockopt(
			das->sock,
			SOL_SOCKET,
			SO_BROADCAST,
			&true,
			sizeof( true )) == -1)
	{
		/*	No.  Throw socket option exception.	*/
		rb_raise(
			eUdpSocket,
			"UdpSocket_Initialize (%s:%d) - Broadcast option setting error,\n"
			"   %s\n",
			__FILE__,
			__LINE__,
			strerror(
				errno ) );
	}

	/*	Set to send and listen on the broadcast address.	*/
	true = 1;
	if (
		bind(
			das->sock,
			(struct sockaddr *) &das->address,
			sizeof( internetAddress )) == -1)
	{
		/*	Throw socket option exception.	*/
		rb_raise(
			eUdpSocket,
			"UdpSocket_Initialize (%s:%d) - Socket bind error\n"
			"   to (%s at %d), %s\n",
			__FILE__,
			__LINE__,
			address,
			das->port,
			strerror(
				errno ) );
	}

	/*	Return result of super send.	*/
	return self;
}

/*-----------------------------------------------------------------------------
| Method:
|	WriteMessage
|
| Parameters:
|	self     - The pointer to the this Ruby object.
|	rPort    - Receiver port.
|	rAddress - Receiver address.
|	rMessage - Message buffer.
|
| Description:
|	Write a broadcast message to a given port in a given broadcast address.
|
|	The first two parameters must be port and address, respectively,
|	followed by the message to write.
|
| Returns:
|	Returns super result.
|
| Raises:
|	ExceptionUdpSocket on write and DNS name resolution errors.
-----------------------------------------------------------------------------*/

static
VALUE
UdpSocket_WriteMessage(
	VALUE	self,
	VALUE	rPort,
	VALUE	rAddress,
	VALUE	rMessage )
{
	auto	int					size;
	auto	unsigned short		port;
	auto	long				length;
	auto	UdpSocketDAs_t		*das;
	auto	long				address;
	auto	struct sockaddr_in	internetAddress;
	auto	struct hostent		*hostEntry;
	auto	uint8_t				*buffer;
	auto	char				*addressString;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		UdpSocketDAs_t,
		das );

	/*	Get address to write to.	*/
	addressString =
		rb_str2cstr(
			rAddress,
			&length );
	port          =
		INT2NUM(
			rPort );

	/*	Can we resolve interface DNS name to address?	*/
	if ( (hostEntry =
		gethostbyname(
			addressString )) == NULL)
	{
		/*	No.  Raise DNS resolution error.	*/
		rb_raise(
			eUdpSocket,
			"UdpSocket_WriteMessage (%s:%d) - DNS resolution error\n"
			"   for (%s), %s\n",
			__FILE__,
			__LINE__,
			addressString,
			strerror(
				errno ) );
	}

	/*	Get address and port number.	*/
	address = *((long *) hostEntry->h_addr_list[ 0 ]);
	port    =
		htons(
			(unsigned short)
			port );

	/*	Set up an AF_INET style address.	*/
	internetAddress.sin_family      = AF_INET;
	internetAddress.sin_port        = port;
	internetAddress.sin_addr.s_addr = address;

	/*	Get buffer to write.	*/
	buffer =
		rb_str2cstr(
			rMessage,
			&length );

	/*	Is the message to large?	*/
	if (length > MAX_MESSAGE)
		rb_raise(
			eUdpSocket,
			"UdpSocket_WriteMessage (%s:%d) - Message to large error,\n"
			"   message is %d bytes\n",
			__FILE__,
			__LINE__,
			length );

	/*	Can we write the message?	*/
	if ( (size =
		sendto(
			das->sock,
			buffer,
			length,
			0,
			(struct sockaddr *) &internetAddress,
			sizeof( internetAddress ) )) < 0)
	{
		/*	No.  Raise write error.	*/
		rb_raise(
			eUdpSocket,
			"UdpSocket_WriteMessage (%s:%d) - Socket sendto error\n"
			"   %s\n",
			__FILE__,
			__LINE__,
			strerror(
				errno ) );
	}

	/*	Return length of message in bytes.	*/
	return
		INT2NUM(
			size );
}

/*-----------------------------------------------------------------------------
| Method:
|	ReadMessage
|
| Parameters:
|	self - The pointer to the this Ruby object.
|
| Description:
|	Read a message received on the UDP socket.
|
| Returns:
|	Returns the message read from the socket.
|
| Raises:
-----------------------------------------------------------------------------*/

static
VALUE
UdpSocket_ReadMessage(
	VALUE	self )
{
	auto	int					size;
	auto	socklen_t			addressSize;
	auto	struct sockaddr_in	internetAddress;
	auto	UdpSocketDAs_t		*das;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		UdpSocketDAs_t,
		das );

	/*	Can we read the message from the socket?	*/
	addressSize = sizeof( internetAddress );
	if ( (size =
		recvfrom(
			das->sock,
			&das->message,
			MAX_MESSAGE,
			0,
			(struct sockaddr *) &internetAddress,
			&addressSize )) < 0)
	{
		/*	No.  Raise read error.	*/
		rb_raise(
			eUdpSocket,
			"UdpSocket_ReadMessage (%s:%d) - Socket recvfrom error\n"
			"   %s\n",
			__FILE__,
			__LINE__,
			strerror(
				errno ) );
	}

	/*	Return message.	*/
	return
		rb_str_new(
			das->message,
			size );
}

/*-----------------------------------------------------------------------------
| Method:
|	Close
|
| Parameters:
|	self - The pointer to the this Ruby object.
|
| Description:
|	Close the listening socket.
|
| Returns:
|	Returns nil.
|
| Raises:
|	Accept errors.
-----------------------------------------------------------------------------*/

static
VALUE
UdpSocket_Close(
	VALUE	self)
{
	auto	UdpSocketDAs_t	*das;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		UdpSocketDAs_t,
		das );

	/*	Close socket.	*/
	if (das->sock >= 0)
	{
		/*	Close the socket.	*/
		close(
			das->sock );
		das->sock = -1;
	}

	/*	Return nil.	*/
	return Qnil;
}

/*-----------------------------------------------------------------------------
| Method:
|	FileNo
|
| Parameters:
|	self - The pointer to the this Ruby object.
|
| Description:
|	Return the socket file descriptor.
|
| Returns:
|	Returns the file number
|
| Raises:
|	None.
-----------------------------------------------------------------------------*/

static
VALUE
UdpSocket_FileNo(
	VALUE	self)
{
	auto	UdpSocketDAs_t		*das;

	/*	Get data attributes.	*/
	Data_Get_Struct(
		self,
		UdpSocketDAs_t,
		das );

	/*	Return file number.	*/
	return
		INT2NUM(
			das->sock );
}

void
Init_UdpSocket(
	void )
{
	/*	Create UdpSocket class.	*/
	cUdpSocket =
		rb_define_class(
			"UdpSocket",
			rb_cObject );

	/*	Add allocation method.	*/
	rb_define_alloc_func(
		cUdpSocket,
		UdpSocket_Allocate );

	/*	Define methods in class.	*/
	rb_define_method(
		cUdpSocket,
		"initialize",
		UdpSocket_Initialize,
		2 );
	rb_define_method(
		cUdpSocket,
		"close",
		UdpSocket_Close,
		0 );
	rb_define_method(
		cUdpSocket,
		"fileno",
		UdpSocket_FileNo,
		0 );
	rb_define_method(
		cUdpSocket,
		"writeMessage",
		UdpSocket_WriteMessage,
		3 );
	rb_define_method(
		cUdpSocket,
		"readMessage",
		UdpSocket_ReadMessage,
		0 );

	/*	Create exception class.	*/
	eUdpSocket =
		rb_define_class(
			"ExceptionUdpSocket",
			rb_eStandardError );
}
